Package: ggfittext
Title: Fit Text Inside a Box in 'ggplot2'
Version: 0.10.4
Authors@R: c(person("David", "Wilkins", email = "david@wilkox.org",
                  role = c("aut", "cre")), 
             person("Brady", "Johnston", email = "brady.johnston@me.com", 
                  role = c("ctb"))
             )
Description: A 'ggplot2' extension to fit text into a box by growing, shrinking or wrapping the text.
Depends: R (>= 3.6)
License: GPL-2
LazyData: true
Imports: grid (>= 3.1), stringi (>= 1.1.2), shades (>= 1.3.1), gridtext
        (>= 0.1.4), ggplot2 (>= 2.2.1), cli (>= 3.6.1)
URL: https://wilkox.org/ggfittext/
BugReports: https://github.com/wilkox/ggfittext/issues/
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), vdiffr, spelling
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-GB
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-05 01:45:37 UTC; wilkox
Author: David Wilkins [aut, cre],
  Brady Johnston [ctb]
Maintainer: David Wilkins <david@wilkox.org>
Repository: CRAN
Date/Publication: 2026-07-05 06:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(getfontsize,richtext_grob)
S3method(getfontsize,text)
S3method(getlabel,richtext_grob)
S3method(getlabel,text)
S3method(getrot,richtext_grob)
S3method(getrot,text)
S3method(makeContent,fittexttree)
S3method(makeContent,fittexttreepolar)
S3method(setfontsize,richtext_grob)
S3method(setfontsize,text)
S3method(setlabel,richtext_grob)
S3method(setlabel,text)
S3method(setx,richtext_grob)
S3method(setx,text)
S3method(sety,richtext_grob)
S3method(sety,text)
S3method(wraplabel,richtext_grob)
S3method(wraplabel,text)
export(geom_bar_text)
export(geom_fit_text)
importFrom(grid,makeContent)

<!-- badges: start -->

[![R-CMD-check](https://github.com/wilkox/ggfittext/workflows/R-CMD-check/badge.svg)](https://github.com/wilkox/ggfittext/actions)
[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/ggfittext)](https://cran.r-project.org/package=ggfittext)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

# ggfittext

ggfittext is a ggplot2 extension for fitting text into boxes.

![](man/figures/README-hero-1.png)<!-- -->

## Installation

Install the release version of ggfittext from CRAN:

``` r
install.packages("ggfittext")
```

If you want the development version, install it from GitHub:

``` r
devtools::install_github("wilkox/ggfittext")
```

## Fitting text inside a box

Sometimes you want to draw some text in a ggplot2 plot so that it fits
inside a defined area. You can do this by manually fiddling with the
font size, but this is tedious and un-reproducible. ggfittext provides a
geom called `geom_fit_text()` that automatically resizes text to fit
inside a box. It works like this:

``` r
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text()
```

![](man/figures/README-unnamed-chunk-4-1.png)<!-- -->

As with `geom_text()`, the position of the text is set by the `x` and
`y` aesthetics. `geom_fit_text()` tries to infer the width and height of
the box in which the text is allowed to fit, and shrinks down any text
that is too big.

## Reflowing text

Another way to make the text fit in the box is by reflowing it; that is,
wrapping it over multiple lines. With the `reflow = TRUE` argument,
`geom_fit_text()` will reflow the text before (if still necessary)
shrinking it:

``` r
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE)
```

![](man/figures/README-unnamed-chunk-5-1.png)<!-- -->

## Growing text

If you want the text to be as large as possible, the argument
`grow = TRUE` will increase the text size to the maximum that will fit
in the box. This works well in conjunction with `reflow`:

``` r
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE, grow = TRUE)
```

![](man/figures/README-unnamed-chunk-6-1.png)<!-- -->

## Placing text

By default, text is placed in the centre of the box. However, you can
place it in a corner or on a side of the box with the `place` argument,
which takes values like “top”, “topright”, “bottomleft” and so on:

``` r
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(place = "topleft", reflow = TRUE)
```

![](man/figures/README-unnamed-chunk-7-1.png)<!-- -->

## Bar plots

ggfittext also provides a geom `geom_bar_text()` for labelling bars in
bar plots:

``` r
ggplot(altitudes, aes(x = craft, y = altitude, label = altitude)) +
  geom_col() +
  geom_bar_text()
```

![](man/figures/README-unnamed-chunk-8-1.png)<!-- -->

`geom_bar_text()` works with stacked bar plots:

``` r
ggplot(beverages, aes(x = beverage, y = proportion, label = ingredient,
                    fill = ingredient)) +
  geom_col(position = "stack") +
  geom_bar_text(position = "stack", reflow = TRUE)
```

![](man/figures/README-unnamed-chunk-9-1.png)<!-- -->

And it works with dodged bar plots, and with flipped bar plots:

``` r
ggplot(beverages, aes(x = beverage, y = proportion, label = ingredient,
                    fill = ingredient)) +
  geom_col(position = "dodge") +
  geom_bar_text(position = "dodge", grow = TRUE, reflow = TRUE, 
                place = "left") +
  coord_flip()
```

![](man/figures/README-unnamed-chunk-10-1.png)<!-- -->

### Experimental feature: rich text

With the `rich = TRUE` argument, `geom_fit_text()` and `geom_bar_text()`
both support a limited subset of Markdown and HTML markup for text
(rendered with [gridtext](https://wilkelab.org/gridtext/)).

``` r
ggplot(animals_rich, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE, grow = TRUE, rich = TRUE)
```

![](man/figures/README-unnamed-chunk-11-1.png)<!-- -->

Rich text cannot be drawn in polar coordinates. Please note that this
feature is liable to change, and is subject to upstream changes to
gridtext.

## Specifying the box limits

If you want to manually set the limits of the box (instead of having
them inferred from `x` and `y`), you can use `xmin` & `xmax` and/or
`ymin` & `ymax`:

``` r
ggplot(presidential, aes(ymin = start, ymax = end, x = party, label = name)) +
  geom_fit_text(grow = TRUE) +
  geom_errorbar(alpha = 0.5)
```

![](man/figures/README-unnamed-chunk-12-1.png)<!-- -->

Alternatively, you can set the width and/or height with the `width`
and/or `height` arguments, which should be `grid::unit()` objects. The
horizontal and/or vertical centre of the box will be defined by `x`
and/or `y`.

## Experimental feature: text in polar coordinates

Text can be drawn in polar coordinates with `geom_fit_text()` simply by
adding `coord_polar()` to the plot. This feature is experimental and any
bug reports are very welcome.

``` r
p <- ggplot(gold, aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, 
                 fill = linenumber, label = line)) +
  coord_polar() +
  geom_rect() +
  scale_fill_gradient(low = "#fee391", high = "#238443")

p + geom_fit_text(min.size = 0, grow = TRUE)
```

![](man/figures/README-unnamed-chunk-13-1.png)<!-- -->

When text is drawn in polar coordinates, the `flip = TRUE` argument can
be used to flip upside-down text the ‘right way up’ to ease readability:

``` r
p + geom_fit_text(min.size = 0, grow = TRUE, flip = TRUE)
```

![](man/figures/README-unnamed-chunk-14-1.png)<!-- -->

## Other useful arguments

All arguments to `geom_fit_text()` can also be used with
`geom_bar_text()`.

- **`contrast`** can be used to automatically invert the colour of the
  text so it contrasts against a background `fill`:

``` r
ggplot(animals, aes(x = type, y = flies, fill = mass, label = animal)) +
  geom_tile() +
  geom_fit_text(reflow = TRUE, grow = TRUE, contrast = TRUE)
```

![](man/figures/README-unnamed-chunk-15-1.png)<!-- -->

- **`padding.x`** and **`padding.y`** can be used to set the padding
  between the text and the edge of the box. By default this is 1 mm.
  These values must be given as `grid::unit()` objects.
- **`min.size`** sets the minimum font size in points, by default 4
  pt. Text smaller than this will be hidden (see also `outside`).
- **`outside`** is `FALSE` by default for `geom_fit_text()`. If `TRUE`,
  text that is placed at “top”, “bottom”, “left” or “right” and must be
  shrunk smaller than `min.size` to fit in the box will be flipped to
  the outside of the box (if it fits there). This is mostly useful for
  drawing text inside bars in a bar plot.
- **`hjust`** and **`vjust`** set the horizontal and vertical
  justification of the text, scaled between 0 (left/bottom) and 1
  (right/top). These are both 0.5 by default.
- **`formatter`** allows you to provide a function that will be applied
  to the text before it is drawn. This is mostly useful in contexts
  where variables may be interpolated, such as when using
  [gganimate](https://gganimate.com/).
- **`fullheight`** is automatically set depending on place, but can be
  overridden with this option. This is used to determine the bounding
  box around the text. If `FALSE`, the bounding box includes the
  x-height of the text and ascenders, but not any descenders. If TRUE,
  it extends from the top of the ascenders to the bottom of the
  descenders. This is mostly useful in situations where you want to
  ensure the baseline of text is consistent between labels
  (`fullheight = FALSE`), or when you want to avoid descenders spilling
  out of the bounding box (`fullheight = TRUE`).

![](man/figures/README-unnamed-chunk-16-1.png)<!-- -->

![](man/figures/README-unnamed-chunk-17-1.png)<!-- -->

## ----fig.width = 4, fig.height = 4--------------------------------------------
library(ggplot2)
library(ggfittext)
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text()

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE, grow = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(place = "topleft", reflow = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(altitudes, aes(x = craft, y = altitude, label = altitude)) +
  geom_col() +
  geom_bar_text()

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(beverages, aes(x = beverage, y = proportion, label = ingredient,
                    fill = ingredient)) +
  geom_col(position = "stack") +
  geom_bar_text(position = "stack", reflow = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(beverages, aes(x = beverage, y = proportion, label = ingredient,
                    fill = ingredient)) +
  geom_col(position = "dodge") +
  geom_bar_text(position = "dodge", grow = TRUE, reflow = TRUE, 
                place = "left") +
  coord_flip()

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(presidential, aes(ymin = start, ymax = end, x = party, label = name)) +
  geom_fit_text(grow = TRUE) +
  geom_errorbar(alpha = 0.5)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals_rich, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE, grow = TRUE, rich = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(gold, aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, 
                 fill = linenumber, label = line)) +
  coord_polar() +
  geom_rect() +
  geom_fit_text(min.size = 0, grow = TRUE) +
  scale_fill_gradient(low = "#fee391", high = "#238443")

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals, aes(x = type, y = flies, fill = mass, label = animal)) +
  geom_tile() +
  geom_fit_text(reflow = TRUE, grow = TRUE, contrast = TRUE)

## ----echo = FALSE, fig.width = 4, fig.height = 0.5----------------------------
ggplot(data.frame(label = "fullheight = TRUE", xmin = 0.25, xmax = 0.75, ymin = 0.25, ymax = 0.75), 
       aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = label)) +
  geom_rect() +
  geom_fit_text(grow = TRUE, fullheight = TRUE, family = "mono") +
  theme(axis.text = element_blank(), axis.ticks = element_blank(), panel.background = element_blank(), plot.title = element_text(hjust = 0.5))

## ----echo = FALSE, fig.width = 4, fig.height = 0.5----------------------------
ggplot(data.frame(label = "fullheight = FALSE", xmin = 0.25, xmax = 0.75, ymin = 0.25, ymax = 0.75), 
       aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = label)) +
  geom_rect() +
  geom_fit_text(grow = TRUE, fullheight = FALSE, family = "mono") +
  theme(axis.text = element_blank(), axis.ticks = element_blank(), panel.background = element_blank(), plot.title = element_text(hjust = 0.5))

#' A 'ggplot2' geom to fit text inside a box
#'
#' `geom_fit_text()` shrinks, grows and wraps text to fit inside a defined
#' box.
#' `geom_bar_text()` is a convenience wrapper around `geom_fit_text()` for
#' labelling bar plots generated with `geom_col()` and `geom_bar()`.
#'
#' @details
#'
#' Except where noted, `geom_fit_text()` behaves more or less like
#' `ggplot2::geom_text()`.
#'
#' There are three ways to define the box in which you want the text to be
#' drawn. The extents of the box on the x and y axes are independent, so any
#' combination of these methods can be used:
#'
#' 1. If the `x` and/or `y` aesthetics are used to set the location of the box,
#' the width and/or height will be set automatically based on the spacing
#' between values on the axis, following the method used by
#' `ggplot2::geom_boxplot()`.
#' 2. Alternatively, if `x` and/or `y` aesthetics are used, the width and/or
#' height of the box can be overridden with a 'width' and/or 'height' argument.
#' These should be `grid::unit()` objects; if not, they will be assumed to use
#' the native axis scale.
#' 3. The boundaries of the box can be set using the aesthetics 'xmin' and
#' 'xmax', and/or 'ymin' and 'ymax'.
#'
#' If the text is too big for the box, it will be shrunk to fit the box. With
#' `grow = TRUE`, the text will be made to fill the box completely whether that
#' requires shrinking or growing.
#'
#' `reflow = TRUE` will cause the text to be reflowed (wrapped) to better fit
#' in the box. If the text cannot be made to fit by reflowing alone, it will be
#' reflowed then shrunk to fit the box. Existing line breaks in the text will
#' be respected when reflowing.
#'
#' `geom_fit_text()` includes experimental support for drawing text in polar
#' coordinates (by adding `coord_polar()` to the plot), however not all
#' features are available when doing so.
#'
#' @section Aesthetics:
#'
#' - label (required)
#' - (xmin AND xmax) OR x (required)
#' - (ymin AND ymax) OR y (required)
#' - alpha
#' - angle
#' - colour
#' - family
#' - fill (with `contrast = TRUE`, used to select a contrasting text colour)
#' - fontface
#' - lineheight
#' - size (in points; note that this differs from `ggplot2::geom_text()`,
#' which expresses size in mm)
#'
#' `geom_bar_text()` requires only the `x` and `y` aesthetics, which are used
#' to infer the extent of the bar. If no `label` aesthetic is provided, the
#' bar's value (`y`, or `x` for flipped bars) will be used as the label.
#'
#' @param padding.x,padding.y Horizontal and vertical padding around the text,
#' expressed in `grid::unit()` objects. Both default to 1 mm.
#' @param min.size Minimum font size, in points. Text smaller than this, whether
#' as supplied or after being shrunk to fit the box, will be hidden. Defaults to
#' 4 pt (8 pt for `geom_bar_text()`).
#' @param place Where inside the box to place the text. Default is 'centre';
#' other options are 'topleft', 'top', 'topright', 'right', 'bottomright',
#' 'bottom', 'bottomleft', 'left', and 'center'/'middle' which are both
#' synonyms for 'centre'. For `geom_bar_text()`, will be set heuristically if
#' not specified.
#' @param outside If `TRUE`, text placed in one of 'top', 'right', 'bottom' or
#' 'left' that would need to be shrunk smaller than `min.size` to fit the box
#' will be drawn outside the box if possible. This is mostly useful for drawing
#' text inside bar/column geoms. Defaults to `TRUE` for `position = "identity"`
#' when using `geom_bar_text()`, otherwise `FALSE`.
#' @param grow If `TRUE`, text will be grown as well as shrunk to fill the box.
#' Defaults to `FALSE`.
#' @param reflow If `TRUE`, text will be reflowed (wrapped) to better fit the
#' box. Defaults to `FALSE`.
#' @param hjust,vjust Horizontal and vertical justification of the text.
#' `vjust` defaults to 0.5. `hjust` defaults to 0.5, except for left placements
#' ('left', 'topleft' and 'bottomleft'), where it defaults to 0, and right
#' placements, where it defaults to 1.
#' @param fullheight If `TRUE`, descenders will be counted when resizing and
#' placing text; if `FALSE`, only the x-height and ascenders will be counted.
#' The main use for this option is for aligning text at the baseline (`FALSE`)
#' or preventing descenders from spilling outside the box (`TRUE`). By default
#' this is set automatically depending on `grow`.
#' @param width,height When using `x` and/or `y` aesthetics, these set the
#' width and/or height of the box. These should be either `grid::unit()`
#' objects or numeric values on the `x` and `y` scales.
#' @param formatter A function that will be applied to the text before it is
#' drawn. This is useful when using `geom_fit_text()` in contexts involving
#' interpolated variables, such as with the 'gganimate' package. `formatter`
#' will be applied serially to each element in the `label` column, so it does
#' not need to be a vectorised function.
#' @param contrast If `TRUE` and in combination with a `fill` aesthetic, the
#' colour of the text will be inverted for better contrast against dark
#' background fills. `FALSE` by default for `geom_fit_text()`, set
#' heuristically for `geom_bar_text()`.
#' @param mapping `ggplot2::aes()` object as standard in 'ggplot2'. Note
#' that aesthetics specifying the box must be provided. See Details.
#' @param rich If `TRUE`, text will be formatted with markdown and HTML markup
#' as implemented by `gridtext::richtext_grob()`. `FALSE` by default. Rich text
#' cannot be drawn in polar coordinates. Please note that rich text support is
#' **experimental** and breaking changes are likely.
#' @param data,stat,position,na.rm,show.legend,inherit.aes,... Standard geom
#' arguments as for `ggplot2::geom_text()`.
#' @param flip If `TRUE`, when in polar coordinates 'upside-down' text will be
#' flipped the 'right way up', to enhance readability.
#'
#' @examples
#'
#' ggplot2::ggplot(ggplot2::presidential, ggplot2::aes(ymin = start, ymax = end,
#'     label = name, x = party)) +
#'   geom_fit_text(grow = TRUE)
#'
#' @rdname geom_fit_text
#' @export
geom_fit_text <- function(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  padding.x = grid::unit(1, "mm"),
  padding.y = grid::unit(1, "mm"),
  min.size = 4,
  place = "centre",
  outside = FALSE,
  grow = FALSE,
  reflow = FALSE,
  hjust = NULL,
  vjust = NULL,
  fullheight = NULL,
  width = NULL,
  height = NULL,
  formatter = NULL,
  contrast = FALSE,
  flip = FALSE,
  rich = FALSE,
  ...
) {
  ggplot2::layer(
    geom = GeomFitText,
    mapping = mapping,
    data = data,
    stat = stat,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      na.rm = na.rm,
      padding.x = padding.x,
      padding.y = padding.y,
      place = place,
      outside = outside,
      min.size = min.size,
      grow = grow,
      reflow = reflow,
      hjust = hjust,
      vjust = vjust,
      fullheight = fullheight,
      width = width,
      height = height,
      formatter = formatter,
      contrast = contrast,
      flip = flip,
      rich = rich,
      ...
    )
  )
}

#' GeomFitText
#' @noRd
GeomFitText <- ggplot2::ggproto(
  "GeomFitText",
  ggplot2::Geom,
  required_aes = c("label"),
  default_aes = ggplot2::aes(
    alpha = 1,
    angle = 0,
    colour = "black",
    family = "",
    fontface = 1,
    lineheight = 0.9,
    size = 12,
    x = NULL,
    y = NULL,
    xmin = NULL,
    xmax = NULL,
    ymin = NULL,
    ymax = NULL,
    fill = NULL
  ),

  setup_params = function(data, params) {
    # Standardise the place argument
    params$place <- ifelse(
      params$place %in% c("middle", "center"),
      "centre",
      params$place
    )

    params
  },

  setup_data = function(
    data,
    params
  ) {
    # Check that valid aesthetics have been supplied for each dimension
    if (!(!is.null(data$xmin) & !is.null(data$xmax) | !is.null(data$x))) {
      cli::cli_abort("geom_fit_text needs either 'xmin' and 'xmax', or 'x'")
    }
    if (!(!is.null(data$ymin) & !is.null(data$ymax) | !is.null(data$y))) {
      cli::cli_abort("geom_fit_text needs either 'ymin' and 'ymax', or 'y'")
    }

    # If 'width' is provided, but not as unit, interpret it as a numeric on the
    # x scale
    if ((!is.null(params$width)) & (!inherits(params$width, "unit"))) {
      data$xmin <- data$x - params$width / 2
      data$xmax <- data$x + params$width / 2
    }

    # If 'height' is provided, but not a unit, interpret it as a numeric on the
    # y scale
    if ((!is.null(params$height)) & (!inherits(params$height, "unit"))) {
      data$ymin <- data$y - params$height / 2
      data$ymax <- data$y + params$height / 2
    }

    # If neither a 'width' parameter nor xmin/xmax aesthetics have been
    # provided, infer the width using the method of geom_boxplot
    if (is.null(params$width) & is.null(data$xmin)) {
      data$width <- ggplot2::resolution(data$x, FALSE) * 0.9
      data$xmin <- data$x - data$width / 2
      data$xmax <- data$x + data$width / 2
      data$width <- NULL
    }

    # If neither a 'height' parameter nor ymin/ymax aesthetics have been
    # provided, infer the height using the method of geom_boxplot
    if (is.null(params$height) & is.null(data$ymin)) {
      data$height <- ggplot2::resolution(data$y, FALSE) * 0.9
      data$ymin <- data$y - data$height / 2
      data$ymax <- data$y + data$height / 2
      data$height <- NULL
    }

    # Apply a formatter function, if one was given
    if (!is.null(params$formatter)) {
      # Check that 'formatter' is a function
      if (!is.function(params$formatter)) {
        cli::cli_abort("`formatter` must be a function")
      }

      # Apply formatter to the labels, checking that the output is a character
      # vector of the correct length
      formatted_labels <- vapply(
        data$label,
        params$formatter,
        character(1),
        USE.NAMES = FALSE
      )
      if (
        (!length(formatted_labels) == length(data$label)) |
          (!is.character(formatted_labels))
      ) {
        cli::cli_abort(
          "`formatter` must produce a character vector of same length as input"
        )
      }
      data$label <- formatted_labels
    }

    data
  },

  draw_key = ggplot2::draw_key_label,

  draw_panel = function(
    data,
    panel_scales,
    coord,
    padding.x = grid::unit(1, "mm"),
    padding.y = grid::unit(1, "mm"),
    min.size = 4,
    grow = FALSE,
    reflow = FALSE,
    hjust = NULL,
    vjust = NULL,
    fullheight = NULL,
    width = NULL,
    height = NULL,
    formatter = NULL,
    contrast = FALSE,
    place = "centre",
    outside = FALSE,
    rich = FALSE,
    flip = FALSE
  ) {
    # Rich text cannot be used in polar coordinates
    if (inherits(coord, "CoordPolar") & rich) {
      cli::cli_abort("Cannot draw rich text in polar coordinates")
    }

    # Transform data to plot scales; if in polar coordinates, we need to ensure
    # that x and y values are given
    if (inherits(coord, "CoordPolar")) {
      if (is.null(data$x)) {
        data$x <- 1
      }
      if (is.null(data$y)) data$y <- 1
    }
    data <- coord$transform(data, panel_scales)

    # For polar coordinates, we need to transform xmin/xmax & ymin/ymax into
    # theta and r values respectively; these can be used later to accurately
    # set the width and height of the bounding box in polar space
    if (inherits(coord, "CoordPolar")) {
      if (!is.null(data$xmin)) {
        data$xmin <- theta_rescale(coord, data$xmin, panel_scales)
      }
      if (!is.null(data$xmax)) {
        data$xmax <- theta_rescale(coord, data$xmax, panel_scales)
      }
      if (!is.null(data$ymin)) {
        data$ymin <- r_rescale(coord, data$ymin, panel_scales$r.range)
      }
      if (!is.null(data$ymax)) {
        data$ymax <- r_rescale(coord, data$ymax, panel_scales$r.range)
      }
    }

    gt <- grid::gTree(
      data = data,
      padding.x = padding.x,
      padding.y = padding.y,
      place = place,
      outside = outside,
      min.size = min.size,
      grow = grow,
      reflow = reflow,
      hjust = hjust,
      vjust = vjust,
      fullheight = fullheight,
      width = width,
      height = height,
      contrast = contrast,
      rich = rich,
      flip = flip,
      cl = ifelse(
        inherits(coord, "CoordPolar"),
        "fittexttreepolar",
        "fittexttree"
      )
    )
    gt$name <- grid::grobName(gt, "geom_fit_text")
    gt
  }
)

#' @importFrom grid makeContent
#' @export
makeContent.fittexttree <- function(x) {
  # Extract data
  ftt <- x
  data <- ftt$data

  # Set and check default parameters
  ftt$outside <- ftt$outside %||% FALSE
  ftt$contrast <- ftt$contrast %||% FALSE
  ftt$vjust <- ftt$vjust %||% 0.5
  if (is.null(ftt$hjust)) {
    if (ftt$place %in% c("left", "bottomleft", "topleft")) {
      ftt$hjust <- 0
    } else if (ftt$place %in% c("right", "bottomright", "topright")) {
      ftt$hjust <- 1
    } else {
      ftt$hjust <- 0.5
    }
  }
  ftt$fullheight <- ftt$fullheight %||% ftt$grow
  ftt$rich <- ftt$rich %||% FALSE

  # Convert padding.x and padding.y to npc units
  ftt$padding.x <- wunit2npc(ftt$padding.x)
  ftt$padding.y <- hunit2npc(ftt$padding.y)

  # Default values for missing aesthetics
  data$fill <- data$fill %||% "grey35"
  data$colour <- data$colour %||% "black"

  # If xmin/xmax are not provided, generate boundary box from width
  if (is.null(data$xmin)) {
    data$xmin <- data$x - (wmm2npc(ftt$width) / 2)
    data$xmax <- data$x + (wmm2npc(ftt$width) / 2)
  }

  # If ymin/ymax are not provided, generate boundary box from height
  if (is.null(data$ymin)) {
    data$ymin <- data$y - (hmm2npc(ftt$height) / 2)
    data$ymax <- data$y + (hmm2npc(ftt$height) / 2)
  }

  # Remove any rows with NA boundaries
  na_rows <- which(
    is.na(data$xmin) | is.na(data$xmax) | is.na(data$ymin) | is.na(data$ymax)
  )
  if (length(na_rows) > 0) {
    data <- data[-na_rows, ]
    cli::cli_warn(
      "Removed {length(na_rows)} rows where box limits were outside plot limits ({.fun geom_fit_text})."
    )
  }

  # Remove any rows with blank labels
  data <- data[!is.na(data$label) & data$label != "", ]

  # Clean up angles
  data$angle <- data$angle %% 360

  # If contrast is set, and the shade of a text colour is too close to the
  # shade of the fill colour, change the colour to its complement
  if (ftt$contrast) {
    # If any fill value is NA, emit a warning and set to the default
    # ggplot2 background grey
    if (!is.null(data$fill)) {
      if (any(is.na(data$fill))) {
        cli::cli_warn("NA values in fill")
        data$fill <- data$fill %NA% "grey35"
      }
    }

    # If any colour value is NA, set to black
    data$colour <- data$colour %NA% "black"

    # Change the text colour to its complement if the background fill is too
    # dark, then change (perhaps again) if the shades of the colour and fill
    # are too similar
    data$colour <- ifelse(
      shades::lightness(data$fill) < 50,
      as.character(shades::complement(shades::shade(data$colour))),
      data$colour
    )
    data$colour <- ifelse(
      abs(shades::lightness(data$fill) - shades::lightness(data$colour)) < 50,
      as.character(shades::complement(shades::shade(data$colour))),
      data$colour
    )
  }

  # Prepare grob for each text label
  grobs <- lapply(seq_len(nrow(data)), function(i) {
    # Convenience
    text <- data[i, ]

    # Get dimensions of bounding box, in mm
    xdim <- wnpc2mm(abs(text$xmin - text$xmax) - (2 * ftt$padding.x))
    ydim <- hnpc2mm(abs(text$ymin - text$ymax) - (2 * ftt$padding.y))

    # Reflow and/or resize the text into a textGrob
    tg <- reflow_and_resize(
      text,
      ftt$reflow,
      ftt$grow,
      ftt$fullheight,
      xdim,
      ydim,
      ftt$rich
    )

    # If the font size is too small and 'outside' has been set, try reflowing
    # and resizing again in the 'outside' position
    if (getfontsize(tg) < ftt$min.size & ftt$outside) {
      if (ftt$place == "top") {
        text$ymin <- text$ymax
        text$ymax <- 1
        ftt$place <- "bottom"
      } else if (ftt$place == "bottom") {
        text$ymax <- text$ymin
        text$ymin <- 0
        ftt$place <- "top"
      } else if (ftt$place == "right") {
        text$xmin <- text$xmax
        text$xmax <- 1
        ftt$place <- "left"
      } else if (ftt$place == "left") {
        text$xmax <- text$xmin
        text$xmin <- 0
        ftt$place <- "right"
      }
      xdim <- wnpc2mm(abs(text$xmin - text$xmax) - (2 * ftt$padding.x))
      ydim <- hnpc2mm(abs(text$ymin - text$ymax) - (2 * ftt$padding.y))
      tg <- setfontsize(tg, text$size)
      ftt$outside <- FALSE
      # If we're moving the text outside and contrast is true, set the text
      # in contrast to the default theme_grey panel colour
      if (ftt$contrast) {
        bg_colour <- "grey92"
        text_colour <- text$colour %||% "black"
        if (
          abs(shades::lightness(bg_colour) - shades::lightness(text_colour)) <
            50
        ) {
          complement <- shades::complement(shades::shade(text_colour))
          text$colour <- as.character(complement)
        }
      }
      tg <- reflow_and_resize(
        text,
        ftt$reflow,
        ftt$grow,
        ftt$fullheight,
        xdim,
        ydim,
        ftt$rich
      )
    }

    # If the font size is still too small, don't draw this label
    if (getfontsize(tg) < ftt$min.size) {
      return()
    }

    # Set hjust and vjust
    tg$hjust <- ftt$hjust
    tg$vjust <- ftt$vjust

    # Update the textGrob dimensions
    tgdim <- tgDimensions(tg, ftt$fullheight, text$angle)

    # To calculate the vector from the geometric centre of the text to the
    # anchor point, we first need the dimensions of the unrotated text in mm.
    # For the common use case of an orthogonal rotation, we can reuse the
    # pre-calculated values to save time
    if (getrot(tg) == 0 | getrot(tg) == 180) {
      tg_width_unrot <- tgdim$width
      tg_height_unrot <- tgdim$height
      if (ftt$fullheight) tg_descent_unrot <- hunit2mm(tgdim$descent)
    } else if (getrot(tg) == 90 | getrot(tg) == 270) {
      tg_width_unrot <- tgdim$height
      tg_height_unrot <- tgdim$width
      if (ftt$fullheight) tg_descent_unrot <- wunit2mm(tgdim$descent)
    } else {
      # For some reason, we don't get accurate values if we do this with the
      # original textGrob so we create a copy
      unrot <- tg
      unrot$rot <- 0
      tg_width_unrot <- wunit2mm(grid::grobWidth(unrot))
      tg_height_unrot <- hunit2mm(grid::grobHeight(unrot))
      if (ftt$fullheight) tg_descent_unrot <- hunit2mm(grid::grobDescent(unrot))
    }

    # We can use these values to calculate the magnitude of the vector from the
    # centre point to the anchor point, using the Pythagorean identity
    if (ftt$fullheight) {
      rise <- ((tg_height_unrot * tg$vjust) + tg_descent_unrot) -
        ((tg_height_unrot + tg_descent_unrot) * 0.5)
    } else {
      rise <- (tg_height_unrot * tg$vjust) - (tg_height_unrot * 0.5)
    }
    run <- (tg_width_unrot * tg$hjust) - (tg_width_unrot * 0.5)
    magnitude <- sqrt((rise^2) + (run^2))

    # The angle between the baseline and the vector can be calculated from the
    # known rise and run
    baseline_angle <- asin(abs(rise) / abs(magnitude)) * (180 / pi)

    # To find the 'direction angle' of the vector (expressed in degrees
    # anti-clockwise from east), we correct for the quadrant then add the
    # rotation of the entire textGrob modulo 360. There's almost certainly a
    # clever trigonometry way to do this but I can't figure it out
    if (sign(rise) == 1 & sign(run) == 1) {
      direction_angle <- baseline_angle
    } else if (sign(rise) == 1 & sign(run) == 0) {
      direction_angle <- 90
    } else if (sign(rise) == 1 & sign(run) == -1) {
      direction_angle <- 180 - baseline_angle
    } else if (sign(rise) == 0 & sign(run) == -1) {
      direction_angle <- 180
    } else if (sign(rise) == -1 & sign(run) == -1) {
      direction_angle <- 180 + baseline_angle
    } else if (sign(rise) == -1 & sign(run) == 0) {
      direction_angle <- 270
    } else if (sign(rise) == -1 & sign(run) == 1) {
      direction_angle <- 360 - baseline_angle
    } else if (sign(rise) == 0 & sign(run) == 1) {
      direction_angle <- 0
    } else if (sign(rise) == 0 & sign(run) == 0) {
      direction_angle <- 0
    }
    direction_angle <- (direction_angle + getrot(tg)) %% 360

    # We can now use these to calculate the x and y offsets of the anchor point
    # from the centre point. For convenience, we will convert these to npc
    x_offset <- wmm2npc(magnitude * cos(deg2rad(direction_angle)))
    y_offset <- hmm2npc(magnitude * sin(deg2rad(direction_angle)))

    # Specify the bounding box limits in npc coordinates
    xmin <- text$xmin + ftt$padding.x
    xmax <- text$xmax - ftt$padding.x
    ymin <- text$ymin + ftt$padding.y
    ymax <- text$ymax - ftt$padding.y

    # Convert the textGrob dimensions into npc
    tgdim$width <- wmm2npc(tgdim$width)
    tgdim$height <- hmm2npc(tgdim$height)

    # Place the text
    if (ftt$place %in% c("topleft", "left", "bottomleft")) {
      x <- xmin + (tgdim$width / 2) + x_offset
    } else if (ftt$place %in% c("top", "centre", "bottom")) {
      x <- ((xmin + xmax) / 2) + x_offset
    } else if (ftt$place %in% c("topright", "right", "bottomright")) {
      x <- xmax - (tgdim$width / 2) + x_offset
    }
    tg <- setx(tg, x)
    if (ftt$place %in% c("topleft", "top", "topright")) {
      y <- ymax - (tgdim$height / 2) + y_offset
    } else if (ftt$place %in% c("left", "centre", "right")) {
      y <- ((ymin + ymax) / 2) + y_offset
    } else if (ftt$place %in% c("bottomleft", "bottom", "bottomright")) {
      y <- ymin + (tgdim$height / 2) + y_offset
    }
    tg <- sety(tg, y)

    # Return the textGrob
    tg
  })

  class(grobs) <- "gList"
  grid::setChildren(ftt, grobs)
}

#' Return a textGrob with the text reflowed and/or resized to fit given
#' dimensions
#'
#' @noRd
reflow_and_resize <- function(
  text,
  reflow,
  grow,
  fullheight,
  xdim,
  ydim,
  rich
) {
  # Create textGrob or richtext_grob
  if (rich) {
    tg <- gridtext::richtext_grob(
      text = text$label,
      x = 0.5,
      y = 0.5,
      hjust = 0.5,
      vjust = 0.5,
      rot = text$angle,
      default.units = "npc",
      gp = grid::gpar(
        col = ggplot2::alpha(text$colour, text$alpha),
        fontsize = text$size,
        fontfamily = text$family,
        fontface = text$fontface,
        lineheight = text$lineheight
      ),
      use_markdown = TRUE
    )
    tg$params <- list(
      text = text$label,
      x = 0.5,
      y = 0.5,
      hjust = 0.5,
      vjust = 0.5,
      rot = text$angle,
      colour = text$colour,
      alpha = text$alpha,
      fontsize = text$size,
      fontfamily = text$family,
      fontface = text$fontface,
      lineheight = text$lineheight
    )
  } else {
    tg <- grid::textGrob(
      label = text$label,
      x = 0.5,
      y = 0.5,
      default.units = "npc",
      hjust = 0.5,
      vjust = 0.5,
      rot = text$angle,
      gp = grid::gpar(
        col = ggplot2::alpha(text$colour, text$alpha),
        fontsize = text$size,
        fontfamily = text$family,
        fontface = text$fontface,
        lineheight = text$lineheight
      )
    )
  }

  # Get starting textGrob dimensions, in mm
  tgdim <- tgDimensions(tg, fullheight, text$angle)

  # Reflow the text, if reflow = TRUE and either the text doesn't currently
  # fit or grow = TRUE and if text contains spaces
  if (
    reflow &
      (grow | tgdim$width > xdim | tgdim$height > ydim) &
      stringi::stri_detect_regex(getlabel(tg), "\\s")
  ) {
    # Generate a list of all wraps for the text
    wraps <- wraplabel(tg)

    # For each wrap, calculate the aspect ratio and return the absolute
    # difference from the box aspect ratio
    wraps$tg <- lapply(wraps$wrap, function(wrap) setlabel(tg, wrap))
    wraps$tgdim <- lapply(
      wraps$tg,
      function(tg) tgDimensions(tg, fullheight, text$angle)
    )
    wraps$aspect_ratio <- vapply(
      wraps$tgdim,
      function(tgdim) tgdim$width / tgdim$height,
      double(1)
    )
    wraps$ar_diff <- abs(wraps$aspect_ratio - (xdim / ydim))

    # If grow is false, select the widest wrap that fits the box (if there is
    # one) and return it
    if (!grow) {
      wraps$width <- unlist(lapply(wraps$tgdim, function(tgdim) tgdim$width))
      wraps$height <- unlist(lapply(wraps$tgdim, function(tgdim) tgdim$height))
      fit_wraps <- wraps[wraps$width < xdim & wraps$height < ydim, ]

      if (nrow(fit_wraps) > 0) {
        best_index <- which(fit_wraps$wrapwidth == max(fit_wraps$wrapwidth))[1]
        tg <- setlabel(tg, fit_wraps$wrap[best_index])
        return(tg)
      }
    }

    # Select the wrap with the smallest difference from the box aspect ratio
    # and update the textGrob dimensions
    best_index <- which(wraps$ar_diff == min(wraps$ar_diff))[1]
    tg <- setlabel(tg, wraps$wrap[best_index])
    tgdim <- wraps$tgdim[best_index][[1]]

    # If the text now fits the bounding box (and we are not trying to grow
    # the text), good to stop here
    if (tgdim$width < xdim & tgdim$height < ydim & !grow) return(tg)
  }

  # Resize text to fit bounding box if it doesn't fit
  if (
    # Standard condition - is text too big for box?
    (tgdim$width > xdim | tgdim$height > ydim) |
      # grow = TRUE condition - is text too small for box?
      (grow & tgdim$width < xdim & tgdim$height < ydim)
  ) {
    # Get the slopes of the relationships between font size and label
    # dimensions
    slopew <- getfontsize(tg) / tgdim$width
    slopeh <- getfontsize(tg) / tgdim$height

    # Calculate the target font size required to fit text to box along each
    # dimension
    targetfsw <- xdim * slopew
    targetfsh <- ydim * slopeh

    # Set to smaller of target font sizes
    tg <- setfontsize(tg, ifelse(targetfsw < targetfsh, targetfsw, targetfsh))
  }

  return(tg)
}

library(ggplot2)
library(vdiffr)

test_that("geom_bar_text() draws test with positions, growing and reflowing", {
  expect_doppelganger("Simple geom_bar_text() plot", {
    ggplot(altitudes, aes(x = craft, y = altitude, label = altitude)) +
      geom_col() +
      geom_bar_text()
  })

  expect_doppelganger("Stacked labels", {
    ggplot(
      beverages,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "stack") +
      geom_bar_text(position = "stack")
  })

  expect_doppelganger("Stacked labels, growing and reflowing", {
    ggplot(
      beverages,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "stack") +
      geom_bar_text(position = "stack", grow = TRUE, reflow = TRUE)
  })

  expect_doppelganger("Dodged labels, growing and reflowing", {
    ggplot(
      beverages,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "dodge") +
      geom_bar_text(
        position = "dodge",
        grow = TRUE,
        reflow = TRUE,
        place = "left"
      )
  })

  expect_doppelganger("Stacked bar plot with '0' values", {
    beverages2 <- beverages
    beverages2[1, 3] <- 0
    ggplot(
      beverages2,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "stack") +
      geom_bar_text(position = "stack", grow = TRUE, reflow = TRUE, size = 48)
  })
})

test_that("geom_bar_text() correctly handles explicit and implied flipped axes", {
  expect_doppelganger("Bar plot with stacked labels and coord_flip()", {
    ggplot(
      beverages,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "stack") +
      geom_bar_text(
        position = "stack",
        grow = TRUE,
        reflow = TRUE,
        place = "left"
      ) +
      coord_flip()
  })

  expect_doppelganger("Bar plot with dodged labels and coord_flip()", {
    ggplot(
      beverages,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "dodge") +
      geom_bar_text(
        position = "dodge",
        grow = TRUE,
        reflow = TRUE,
        place = "left"
      ) +
      coord_flip()
  })

  expect_doppelganger("Bar plot with implied axis flip", {
    ggplot(altitudes, aes(x = altitude, y = craft, label = altitude)) +
      geom_col() +
      geom_bar_text(grow = TRUE)
  })

  expect_doppelganger("Bar plot with coord_flip() and place = middle", {
    ggplot(altitudes, aes(x = craft, y = altitude, label = altitude)) +
      coord_flip() +
      geom_col() +
      geom_bar_text(place = "middle")
  })

  expect_doppelganger("Bar plot with implied axis flip and place = middle", {
    ggplot(altitudes, aes(x = altitude, y = craft, label = altitude)) +
      geom_col() +
      geom_bar_text(place = "middle")
  })

  expect_doppelganger("Integer y shouldn't induce an implied flip", {
    beverages2 <- beverages
    beverages2$proportion <- 1:6
    ggplot(
      beverages2,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "stack") +
      geom_bar_text(position = "stack")
  })
})

test_that("horizontal bars are labelled with the value, not the category index (#45)", {
  p <- ggplot(
    data.frame(cat = c("a", "b"), val = c(10, 20)),
    aes(x = val, y = cat)
  ) +
    geom_col() +
    geom_bar_text()
  expect_equal(layer_data(p, 2)$label, c(10, 20))
})

test_that("geom_bar_text() works with faceting", {
  expect_doppelganger("geom_bar_text() with faceting", {
    ggplot(
      beverages,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "stack") +
      geom_bar_text(
        position = "stack",
        grow = TRUE,
        reflow = TRUE,
        place = "left"
      ) +
      facet_wrap(~beverage)
  })
})

test_that("geom_bar_text() draws contrasting text", {
  expect_doppelganger("Contrast against default bar colour", {
    ggplot(altitudes, aes(x = craft, y = altitude, label = altitude)) +
      geom_col() +
      geom_bar_text(contrast = TRUE)
  })

  expect_doppelganger("Contrasting with non-black text", {
    ggplot(animals, aes(x = type, y = flies, fill = mass, label = animal)) +
      geom_tile() +
      geom_fit_text(
        colour = "thistle",
        reflow = TRUE,
        grow = TRUE,
        contrast = TRUE
      ) +
      scale_fill_gradientn(
        colours = c("red", "yellow", "green", "lightblue", "darkblue"),
        values = c(1.0, 0.8, 0.6, 0.4, 0.2, 0)
      )
  })
})

test_that("contrast auto-detection tolerates NA colours (#56)", {
  # When contrast is left to auto-detect and every non-NA text colour is the
  # default black, a single NA colour previously made all(colour == "black")
  # evaluate to NA, crashing makeContent() with "missing value where
  # TRUE/FALSE needed". Auto-detection should ignore NA colours instead.
  df <- data.frame(
    cat = c("a", "b", "c"),
    val = c(1, 1, 1),
    txt = c("black", "black", NA)
  )
  p <- ggplot(df, aes(x = cat, y = val, label = cat, colour = txt)) +
    geom_col() +
    geom_bar_text() +
    scale_colour_identity()

  pdf(NULL)
  on.exit(dev.off())
  expect_no_error(print(p))
})

test_that("outside default is derived from position (string or object)", {
  # position_identity(): outside should default to TRUE, whether the position
  # is given as a string or as a Position object (#44)
  expect_true(geom_bar_text(position = "identity")$geom_params$outside)
  expect_true(geom_bar_text(position = position_identity())$geom_params$outside)

  # Any other position defaults outside to FALSE. A Position object is required
  # whenever a dodge width must be set, and comparing it with == would error.
  expect_false(geom_bar_text(position = "stack")$geom_params$outside)
  expect_false(
    geom_bar_text(position = position_dodge(width = 0.9))$geom_params$outside
  )

  # An explicit outside is always honoured over the position-derived default.
  expect_false(geom_bar_text(position = "identity", outside = FALSE)$geom_params$outside)
})

# geom_bar_text() nests one fittexttree per sign group inside an outer gTree, so
# we reach into the positive group and draw it directly. The short bar (0.1)
# cannot fit its label inside; the tall bar (100) always can. So the short-bar
# label is dropped when outside = FALSE and drawn above the bar when
# outside = TRUE.
test_that("geom_bar_text() draws short-bar labels outside the bar (#44)", {
  df <- data.frame(craft = c("a", "b"), altitude = c(0.1, 100))
  draw <- function(outside) {
    p <- ggplot(df, aes(x = craft, y = altitude, label = altitude)) +
      geom_col() +
      geom_bar_text(outside = outside, min.size = 8)
    grDevices::pdf(NULL)
    on.exit(grDevices::dev.off())
    outer <- suppressWarnings(layer_grob(p, 2)[[1]])
    # Both bars are positive, so there is a single fittexttree (the positives
    # group). Assert that structure before reaching into it, so a refactor of
    # the nesting fails here rather than silently drawing the wrong group.
    expect_length(outer$children, 1)
    expect_s3_class(outer$children[[1]], "fittexttree")
    grid::makeContent(outer$children[[1]])$children
  }

  # The tall bar is always labelled; the short bar is only labelled when its
  # label may be drawn outside the bar.
  expect_length(draw(FALSE), 1)
  expect_length(draw(TRUE), 2)
})

# A valid formatter is applied to the labels, including the value-axis labels
# geom_bar_text() infers when no label aesthetic is mapped.
test_that("geom_bar_text() applies a valid formatter to the labels", {
  p <- ggplot(data.frame(x = c("a", "b"), y = c(1, 2)), aes(x, y)) +
    geom_col() +
    geom_bar_text(formatter = function(x) paste0(x, "%"))
  expect_equal(layer_data(p, 2)$label, c("1%", "2%"))
})

# GeomBarText carries its own copy of the formatter validation, parallel to
# GeomFitText's; exercise its error paths independently so the two cannot drift
# apart unnoticed.
test_that("geom_bar_text() rejects invalid formatters", {
  base <- ggplot(data.frame(x = c("a", "b"), y = c(1, 2)), aes(x, y)) +
    geom_col()

  # Not a function at all (the geom's own check).
  expect_error(
    ggplot_build(base + geom_bar_text(formatter = "nope")),
    "must be a function"
  )
  # Wrong type and wrong length are rejected by vapply's template enforcement.
  expect_error(ggplot_build(base + geom_bar_text(formatter = function(x) 1)))
  expect_error(
    ggplot_build(base + geom_bar_text(formatter = function(x) c(x, x)))
  )
})

test_that("width works as expected", {
  expect_doppelganger("geom_bar_text() with width to match geom_col() width", {
    ggplot(
      beverages,
      aes(x = beverage, y = proportion, label = ingredient, fill = ingredient)
    ) +
      geom_col(position = "stack", width = 0.5) +
      geom_bar_text(
        position = "stack",
        reflow = TRUE,
        width = 0.5,
        grow = TRUE,
        fullheight = TRUE
      )
  })
})

library(ggplot2)

test_that("geom_fit_text() plots with rich text", {

  skip_on_os(c("windows", "linux"))

  expect_doppelganger("Basic rich text", {
    ggplot(animals_rich, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(rich = TRUE)
  } )

  expect_doppelganger("Basic rich text with grow", {
   ggplot(animals_rich, aes(x = type, y = flies, label = animal)) +
     geom_tile(fill = "white", colour = "black") +
     geom_fit_text(rich = TRUE, grow = TRUE)
  } )

  expect_doppelganger("Basic rich text with reflow", {
   ggplot(animals_rich, aes(x = type, y = flies, label = animal)) +
     geom_tile(fill = "white", colour = "black") +
     geom_fit_text(rich = TRUE, reflow = TRUE)
  } )

  expect_doppelganger("Basic rich text with grow and reflow", {
   ggplot(animals_rich, aes(x = type, y = flies, label = animal)) +
     geom_tile(fill = "white", colour = "black") +
     geom_fit_text(rich = TRUE, grow = TRUE, reflow = TRUE)
  } )

  animals_rich2 <- animals_rich
  animals_rich2$animal[1] <- paste(
    "Whose **woods these *are* I** < think <sub>I</sub> <span>know",
    "</span>. His <sup>house</sup> **is** in > *the* ***village*** though.",
    "What do ****four asterisks**** do?<br>",
    "What about *****five asterisks*****?",
    "This is < less <sub>than</sub> <sup>this</sup>, and a * couple of ** floating asterisks.",
    sep = " "
  )
  animals_rich2$animal[6] <- paste(
    "<span style='color:red'",
    ">",
    "red</span>-rumped parrot",
    sep = " "
  )

  expect_doppelganger("Complex rich text with grow and reflow", {
    ggplot(animals_rich2, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(rich = TRUE, grow = TRUE, reflow = TRUE)
  } )

  expect_doppelganger("Complex rich text with reflow only", {
    ggplot(animals_rich2, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(rich = TRUE, reflow = TRUE)
  } )

} )

test_that("geom_bar_text() plots with rich text", {

  skip_on_os(c("windows", "linux"))

  expect_doppelganger("Basic bar plot with rich text", {
    ggplot(beverages_rich, aes(x = beverage, y = proportion, label = ingredient, 
                               fill = ingredient)) +
      geom_col(position = "stack") +
      geom_bar_text(position = "stack", rich = TRUE)
  } )

  expect_doppelganger("Rich bar plot with grow", {
    ggplot(beverages_rich, aes(x = beverage, y = proportion, label = ingredient, 
                               fill = ingredient)) +
      geom_col(position = "stack") +
      geom_bar_text(position = "stack", rich = TRUE, grow = TRUE)
  } )

  expect_doppelganger("Rich bar plot with reflow", {
    ggplot(beverages_rich, aes(x = beverage, y = proportion, label = ingredient, 
                               fill = ingredient)) +
      geom_col(position = "stack") +
      geom_bar_text(position = "stack", rich = TRUE, reflow = TRUE)
  } )

  expect_doppelganger("Rich bar plot with grow and reflow", {
    ggplot(beverages_rich, aes(x = beverage, y = proportion, label = ingredient, 
                               fill = ingredient)) +
      geom_col(position = "stack") +
      geom_bar_text(position = "stack", rich = TRUE, grow = TRUE, reflow = TRUE)
  } )

} )

library(ggplot2)
library(vdiffr)

test_that("geom_fit_text() places text correctly", {
  expect_doppelganger("Default placement (middle)", {
    ggplot(animals, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text()
  })

  expect_doppelganger("Placing at the top", {
    ggplot(animals, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(place = "top")
  })

  expect_doppelganger("Placing at the bottom right", {
    ggplot(animals, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(place = "bottomright")
  })
})

test_that("geom_fit_text() grows and reflows text correctly", {
  expect_doppelganger("Growing without reflowing", {
    ggplot(animals, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(grow = TRUE)
  })

  expect_doppelganger("Growing with placement at the bottom", {
    ggplot(animals, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(grow = TRUE, place = "bottom")
  })

  expect_doppelganger("Growing with reflowing", {
    ggplot(animals, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(grow = TRUE, reflow = TRUE)
  })

  expect_doppelganger("Reflowing without growing", {
    ggplot(animals, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(reflow = TRUE)
  })

  animals2 <- animals
  animals2$animal[1] <- paste(
    "Whose woods these are I think I know.",
    "His house is in the village though;",
    "He will not see me stopping here",
    "To watch his woods fill up with snow.",
    sep = "\n"
  )
  animals2$animal[2] <- paste(
    "Whose woods these are I think I know.",
    "His house is in the village though;",
    "He will not see me stopping here",
    "To watch his woods fill up with snow.",
    sep = " "
  )

  expect_doppelganger("Complex text with reflowing and growing", {
    ggplot(animals2, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(grow = TRUE, reflow = TRUE)
  })

  expect_doppelganger("Complex text with reflowing only", {
    ggplot(animals2, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(reflow = TRUE)
  })
})

test_that("The width and height arguments work correctly", {
  expect_doppelganger("Numeric width and height", {
    df <- data.frame(
      vehicle = c("light plane", "jumbo jet", "space shuttle"),
      x = c(15, 40, 85),
      y = c(15, 85, 23.5)
    )
    ggplot(df, aes(x = x, y = y, label = vehicle)) +
      geom_tile(width = 10, height = 20, fill = "gray") +
      geom_fit_text(width = 10, height = 20)
  })
})

test_that("The formatter argument works correctly", {
  expect_doppelganger("Label with formatter function", {
    wikif <- function(x) {
      paste0(x, " (citation needed)")
    }
    ggplot(animals, aes(x = type, y = flies, label = animal)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(grow = TRUE, reflow = TRUE, formatter = wikif)
  })
})

test_that("geom_fit_text() handles blank labels", {
  expect_doppelganger("Blank label", {
    presidential$name[1] <- ""
    ggplot(
      presidential,
      aes(ymin = start, ymax = end, label = name, x = party)
    ) +
      geom_fit_text(grow = TRUE)
  })
})

# Logic-level test for the blank-label filter (#48). The vdiffr snapshot above
# would still pass with the old, accidentally-working condition, so we inspect
# the drawn grob tree directly: NA and empty labels must be dropped, leaving one
# child grob for the single real label.
test_that("makeContent.fittexttree() drops NA and empty labels (#48)", {
  df <- data.frame(x = 1:3, y = 1:3, label = c("a", NA, ""))
  p <- ggplot(df, aes(x, y, label = label)) + geom_fit_text()

  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off())

  ftt <- suppressWarnings(layer_grob(p)[[1]])
  drawn <- grid::makeContent(ftt)
  expect_length(drawn$children, 1)
})

test_that("geom_fit_text() warns when box limits are outside of plot limits", {
  df <- data.frame(x = letters[1:5], y = 0:4, lb = 3)

  expect_warning(
    {
      p <- ggplot(df, aes(x = x, y = y, label = lb)) +
        geom_fit_text() +
        ylim(0.5, 3)
      print(p)
    },
    "box limits were outside plot limits"
  )

  expect_warning(
    {
      p <- ggplot(df, aes(x = x, y = y, label = lb)) +
        ylim(-0.1, 6) +
        geom_fit_text()
      print(p)
    },
    "box limits were outside plot limits"
  )

  expect_warning(
    {
      p <- ggplot(df, aes(x = y, y = x, label = lb)) +
        xlim(-0.1, 6) +
        geom_fit_text()
      print(p)
    },
    "box limits were outside plot limits"
  )
})

test_that("geom_fit_text() draws angled text correctly", {
  expect_doppelganger("Angled text", {
    ggplot(animals, aes(x = type, y = flies, label = animal, angle = mass)) +
      geom_tile(fill = "white", colour = "black") +
      geom_fit_text(size = 48)
  })
})

test_that("geom_fit_text() draws in boxes defined with xmin/xmax and ymin/ymax", {
  expect_doppelganger("Box defined with x and ymin/ymax", {
    ggplot(
      presidential,
      aes(ymin = start, ymax = end, x = party, label = name)
    ) +
      geom_fit_text(grow = TRUE, size = 48) +
      geom_errorbar(alpha = 0.5)
  })

  expect_doppelganger("Box defined with xmin/xmax and y", {
    ggplot(
      presidential,
      aes(xmin = start, xmax = end, y = party, label = name)
    ) +
      geom_fit_text(grow = TRUE, size = 48) +
      geom_errorbar(alpha = 0.5)
  })

  expect_doppelganger("Box defined with xmin/xmax and ymin/ymax", {
    ggplot(
      presidential,
      aes(xmin = start, xmax = end, ymin = start, ymax = end, label = name)
    ) +
      geom_fit_text(grow = TRUE, size = 48)
  })
})

test_that("geom_fit_text() draws contrasting text", {
  expect_doppelganger("Contrasting text", {
    ggplot(animals, aes(x = type, y = flies, fill = mass, label = animal)) +
      geom_tile() +
      geom_fit_text(reflow = TRUE, grow = TRUE, contrast = TRUE, size = 48)
  })

  expect_doppelganger("Contrasting with non-black text", {
    ggplot(animals, aes(x = type, y = flies, fill = mass, label = animal)) +
      geom_tile() +
      geom_fit_text(
        colour = "thistle",
        reflow = TRUE,
        grow = TRUE,
        contrast = TRUE
      ) +
      scale_fill_gradientn(
        colours = c("red", "yellow", "green", "lightblue", "darkblue"),
        values = c(1.0, 0.8, 0.6, 0.4, 0.2, 0)
      )
  })
})

test_that("geom_fit_text() draws, grows and correctly places text in polar coordinates", {
  expect_doppelganger("Polar geom_fit_text() plot", {
    ggplot(
      gold,
      aes(label = line, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)
    ) +
      geom_rect(fill = "lightblue") +
      coord_polar(start = (pi / 2)) +
      geom_fit_text(min.size = 0)
  })

  expect_doppelganger("Polar plot with growing, non-fullheight", {
    ggplot(
      gold,
      aes(label = line, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)
    ) +
      geom_rect(fill = "lightblue") +
      coord_polar(start = (pi / 2)) +
      geom_fit_text(grow = TRUE, fullheight = FALSE, min.size = 0)
  })

  expect_doppelganger("Polar plot with growing, fullheight", {
    ggplot(
      gold,
      aes(label = line, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)
    ) +
      geom_rect(fill = "lightblue") +
      coord_polar(start = (pi / 2)) +
      geom_fit_text(grow = TRUE, fullheight = TRUE, min.size = 0)
  })

  expect_doppelganger("Polar plot with text placed at the top", {
    ggplot(
      gold,
      aes(label = line, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)
    ) +
      geom_rect(fill = "lightblue") +
      coord_polar(start = (pi / 2)) +
      geom_fit_text(place = "top", min.size = 0)
  })

  expect_doppelganger("Polar plot with text placed at the bottom", {
    ggplot(
      gold,
      aes(label = line, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)
    ) +
      geom_rect(fill = "lightblue") +
      coord_polar(start = (pi / 2)) +
      geom_fit_text(place = "bottom", min.size = 0)
  })
})


test_that("geom_fit_text() emits an error if you try to draw rich text in polar coordinates", {
  expect_error({
    p <- ggplot(
      gold,
      aes(label = line, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)
    ) +
      geom_rect(fill = "lightblue") +
      coord_polar(start = (pi / 2)) +
      geom_fit_text(min.size = 0, rich = TRUE)
    print(p)
  })
})


test_that("geom_fit_text() correctly applies width and height arguments in polar coordinates", {
  expect_doppelganger("Polar plot with height of 20 mm", {
    ggplot(gold) +
      coord_polar() +
      geom_rect(aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)) +
      geom_fit_text(
        aes(label = line, xmin = xmin, xmax = xmax, y = (ymin + ymax) / 2),
        min.size = 0,
        grow = TRUE,
        height = grid::unit(20, "mm")
      )
  })

  expect_doppelganger("Polar plot with height of 5 mm", {
    ggplot(gold) +
      coord_polar() +
      geom_rect(aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)) +
      geom_fit_text(
        aes(label = line, xmin = xmin, xmax = xmax, y = (ymin + ymax) / 2),
        min.size = 0,
        grow = TRUE,
        height = grid::unit(5, "mm")
      )
  })

  expect_doppelganger("Polar plot with width of 100 mm", {
    ggplot(gold) +
      coord_polar() +
      geom_rect(aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)) +
      geom_fit_text(
        aes(label = line, x = (xmin + xmax) / 2, ymin = ymin, ymax = ymax),
        min.size = 0,
        grow = TRUE,
        width = grid::unit(100, "mm")
      )
  })

  expect_doppelganger("Polar plot with width of 40 mm", {
    ggplot(gold) +
      coord_polar() +
      geom_rect(aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax)) +
      geom_fit_text(
        aes(label = line, x = (xmin + xmax) / 2, ymin = ymin, ymax = ymax),
        min.size = 0,
        grow = TRUE,
        width = grid::unit(40, "mm")
      )
  })
})

test_that("The flip argument works properly when drawing in polar coordinates", {
  expect_doppelganger("Polar plot with flip", {
    ggplot(
      gold,
      aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = line)
    ) +
      coord_polar() +
      geom_rect() +
      geom_fit_text(min.size = 0, flip = TRUE, grow = TRUE)
  })

  expect_doppelganger("Polar plot with flip and shifted start", {
    ggplot(
      gold,
      aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = line)
    ) +
      coord_polar(start = (pi / 2)) +
      geom_rect() +
      geom_fit_text(min.size = 0, flip = TRUE, grow = TRUE)
  })

  expect_doppelganger("Polar plot with flip and text in top left", {
    ggplot(
      gold,
      aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = line)
    ) +
      coord_polar() +
      geom_rect() +
      geom_fit_text(min.size = 0, flip = TRUE, place = "topleft")
  })
})

# Logic-level test for the reflow engine. Unlike the vdiffr snapshots above, it
# inspects the returned grob directly, so it catches metric bugs that leave a
# visually plausible but incorrect result.

test_that("reflow_and_resize() never returns an unshrunk wrap that overflows the box (#47)", {

  # A label whose successive candidate wraps are non-monotone in rendered width
  # (mixed wide/narrow glyphs). This is what exposed the bug: the "widest wrap
  # that fits" filter could select a wrap that does not fit, and because that
  # branch returns early the text was never shrunk to compensate.
  label <- "MMMM ii MMMM iiiiiiiiii MM iii MMMMM WW i WWWW"
  text <- data.frame(
    label = label, size = 12, angle = 0, colour = "black", alpha = 1,
    family = "", fontface = 1, lineheight = 0.9, stringsAsFactors = FALSE
  )

  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off())

  ydim <- 30

  # Sweep box widths across the range where candidate wrap widths diverge. The
  # exact trigger band is font-metric-dependent, so we assert the invariant over
  # a broad sweep rather than at one fragile point.
  for (xdim in seq(40, 70, by = 0.5)) {
    tg <- reflow_and_resize(
      text, reflow = TRUE, grow = FALSE, fullheight = FALSE,
      xdim = xdim, ydim = ydim, rich = FALSE
    )

    # If the font size is unchanged, no shrinking was applied, so the chosen
    # wrap must genuinely fit the box in both dimensions.
    if (isTRUE(all.equal(getfontsize(tg), text$size))) {
      dim <- tgDimensions(tg, FALSE, text$angle)
      expect_lte(dim$width, xdim + 1e-6)
      expect_lte(dim$height, ydim + 1e-6)
    }
  }
})

test_that("GeomFitText$setup_data() does not add a dead 'flip' column", {
  p <- ggplot(animals, aes(x = type, y = flies, label = animal)) +
    geom_fit_text()
  expect_false("flip" %in% names(layer_data(p)))
})

test_that("deprecated geom_grow_text()/geom_shrink_text() stubs are removed", {
  ns <- asNamespace("ggfittext")
  expect_false(exists("geom_grow_text", envir = ns, inherits = FALSE))
  expect_false(exists("geom_shrink_text", envir = ns, inherits = FALSE))
})

# Logic-level test for the min.size gate. A label that can only be fitted by
# shrinking below min.size must be dropped entirely, leaving no child grob. With
# grow = FALSE the 12 pt text is never enlarged, so a 200 pt floor is guaranteed
# to reject every label whatever the box size or font metrics.
test_that("labels that must shrink below min.size produce no child grob", {
  df <- data.frame(x = 1:3, y = 1:3, label = c("a", "b", "c"))
  p <- ggplot(df, aes(x, y, label = label)) + geom_fit_text(min.size = 200)

  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off())

  drawn <- grid::makeContent(suppressWarnings(layer_grob(p)[[1]]))
  expect_length(drawn$children, 0)
})

# hjust is derived from the horizontal component of `place`, and manual
# hjust/vjust must override the derived defaults. The box is large enough that
# the label always draws (asserted before indexing, so a dropped label fails
# semantically rather than with a subscript error).
test_that("hjust follows place, and manual hjust/vjust override it", {
  df <- data.frame(xmin = 0, xmax = 10, ymin = 0, ymax = 10, label = "hi")
  child <- function(...) {
    p <- ggplot(
      df,
      aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = label)
    ) +
      geom_fit_text(...)
    grDevices::pdf(NULL)
    on.exit(grDevices::dev.off())
    drawn <- grid::makeContent(suppressWarnings(layer_grob(p)[[1]]))
    expect_length(drawn$children, 1)
    drawn$children[[1]]
  }

  # hjust follows the horizontal component of `place`. vjust is not
  # place-derived — vertical position is handled by the placement maths, so it
  # always defaults to 0.5.
  hjusts <- c(
    topleft = 0, left = 0, bottomleft = 0,
    top = 0.5, centre = 0.5, bottom = 0.5,
    topright = 1, right = 1, bottomright = 1
  )
  for (place in names(hjusts)) {
    tg <- child(place = place)
    expect_equal(tg$hjust, hjusts[[place]], info = place)
    expect_equal(tg$vjust, 0.5, info = place)
  }

  # Explicit hjust/vjust take precedence over the place-derived defaults.
  overridden <- child(place = "centre", hjust = 0, vjust = 1)
  expect_equal(overridden$hjust, 0)
  expect_equal(overridden$vjust, 1)
})

# The `outside` path (R/geom_fit_text.R draw of below-min.size labels) is the
# flagship behaviour of geom_bar_text() but lives in the shared engine. A label
# too big for its box but with room outside it should be dropped when
# outside = FALSE and rescued when outside = TRUE.
test_that("outside = TRUE draws below-min.size labels outside the box", {
  # A wide but extremely short box, with ample room above it in the panel.
  df <- data.frame(xmin = 0, xmax = 100, ymin = 0, ymax = 0.1, label = "hi")
  draw <- function(outside) {
    p <- ggplot(
      df,
      aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = label)
    ) +
      geom_fit_text(place = "top", outside = outside, min.size = 8) +
      ylim(0, 10)
    grDevices::pdf(NULL)
    on.exit(grDevices::dev.off())
    grid::makeContent(suppressWarnings(layer_grob(p)[[1]]))$children
  }

  expect_length(draw(FALSE), 0)
  expect_length(draw(TRUE), 1)
})

# A valid formatter is applied to each label during the build step.
test_that("geom_fit_text() applies a valid formatter to the labels", {
  p <- ggplot(
    data.frame(x = 1:2, y = 1:2, label = c("a", "b")),
    aes(x, y, label = label)
  ) +
    geom_fit_text(formatter = function(x) paste0(x, "!"))
  expect_equal(layer_data(p)$label, c("a!", "b!"))
})

# An invalid formatter must be rejected during the build step, before any
# drawing. A non-function is caught by the geom's own check; a formatter that
# returns the wrong type or length is rejected by vapply's template enforcement
# before that check. Either way a bad formatter never reaches drawing.
test_that("geom_fit_text() rejects invalid formatters", {
  base <- ggplot(
    data.frame(x = 1:2, y = 1:2, label = c("a", "b")),
    aes(x, y, label = label)
  )

  # Not a function at all (the geom's own check).
  expect_error(
    ggplot_build(base + geom_fit_text(formatter = "nope")),
    "must be a function"
  )

  # Returns the wrong type.
  expect_error(ggplot_build(base + geom_fit_text(formatter = function(x) 1)))

  # Returns more than one value per label.
  expect_error(
    ggplot_build(base + geom_fit_text(formatter = function(x) c(x, x)))
  )
})

#' Default values
#'
#' @noRd
"%||%" <- function(a, b) if (is.null(a)) b else a
"%NA%" <- function(a, b) ifelse(is.na(a), b, a)

#' Conversions between degrees and radians
#'
#' @noRd
deg2rad <- function(deg) { deg * (pi / 180) }
rad2deg <- function(rad) { rad * (180 / pi) }

#' Return a reversed string
#' @param string String to be returned with reversed characters.
#' @noRd
strrev <- function(string) {
  lapply(lapply(strsplit(string, NULL), rev), paste0, collapse = "")
}

#' Textgrob dimensions, in mm
#' 
#' @noRd
tgDimensions <- function(tg, fullheight, angle) {
  width <- wunit2mm(grid::grobWidth(tg))
  height <- hunit2mm(grid::grobHeight(tg))
  if (fullheight) {
    descent <- grid::grobDescent(tg)
    width <- width + abs(wunit2mm(descent) * sin(deg2rad(angle)))
    height <- height + abs(hunit2mm(descent) * cos(deg2rad(angle)))
  } else {
    descent <- NULL
  }
  list(width = width, height = height, descent = descent)
}

#' Width and height unit conversions
#'
#' @noRd
wunit2npc <- function(w) grid::convertWidth(w, "npc", valueOnly = TRUE)
wmm2npc <- function(w) wunit2npc(grid::unit(w, "mm"))
wunit2mm <- function(w) grid::convertWidth(w, "mm", valueOnly = TRUE)
wnpc2mm <- function(w) wunit2mm(grid::unit(w, "npc"))
hunit2npc <- function(h) grid::convertHeight(h, "npc", valueOnly = TRUE)
hmm2npc <- function(h) hunit2npc(grid::unit(h, "mm"))
hunit2mm <- function(h) grid::convertHeight(h, "mm", valueOnly = TRUE)
hnpc2mm <- function(h) hunit2mm(grid::unit(h, "npc"))

#' Getters and setters for textGrob and richtext_grob
#'
#' @noRd
set_richtext_grob_param <- function(tg, param, value) {
  params <- tg$params
  params[param] <- value
  tg <- gridtext::richtext_grob(
    text = params$text,
    x = params$x,
    y = params$y,
    hjust = params$hjust,
    vjust = params$vjust,
    rot = params$rot,
    default.units = "npc",
    gp = grid::gpar(
      col = ggplot2::alpha(params$colour, params$alpha),
      fontsize = params$fontsize,
      fontfamily = params$fontfamily,
      fontface = params$fontface,
      lineheight = params$lineheight
    ),
    use_markdown = TRUE
  )
  tg$params <- params
  tg
}

getrot <- function(tg) UseMethod("getrot")

#' @export
getrot.text <- function(tg) tg$rot

#' @export
getrot.richtext_grob <- function(tg) tg$params$rot

getlabel <- function(tg) UseMethod("getlabel")

#' @export
getlabel.text <- function(tg) tg$label

#' @export
getlabel.richtext_grob <- function(tg) tg$params$text

setlabel <- function(tg, value) UseMethod("setlabel")

#' @export
setlabel.text <- function(tg, value) {
  tg$label <- value
  tg
}

#' @export
setlabel.richtext_grob <- function(tg, value) set_richtext_grob_param(tg, "text", value)

getfontsize <- function(tg) UseMethod("getfontsize")

#' @export
getfontsize.text <- function(tg) tg$gp$fontsize

#' @export
getfontsize.richtext_grob <- function(tg) tg$params$fontsize

setfontsize <- function(tg, value) UseMethod("setfontsize")

#' @export
setfontsize.text <- function(tg, value) {
  tg$gp$fontsize <- value
  tg
}

#' @export
setfontsize.richtext_grob <- function(tg, value) {
  set_richtext_grob_param(tg, "fontsize", value)
}

setx <- function(tg, x) UseMethod("setx")

#' @export
setx.text <- function(tg, x) {
  if (!inherits(x, "unit")) x <- grid::unit(x, "npc")
  tg$x <- x
  tg
}

#' @export
setx.richtext_grob <- function(tg, x) {
  set_richtext_grob_param(tg, "x", x)
}

sety <- function(tg, y) UseMethod("sety")

#' @export
sety.text <- function(tg, y) {
  if (!inherits(y, "unit")) y <- grid::unit(y, "npc")
  tg$y <- y
  tg
}

#' @export
sety.richtext_grob <- function(tg, y) {
  tg <- set_richtext_grob_param(tg, "y", y)
}

#' Methods to wrap labels for textGrob and richtext_grob
#'
#' Each method returns a data frame containing all possible wraps for the text.
#' Newline tokens (`\n` for plain text, additionally `<br>` and its variants
#' for rich text) are respected when wrapping.
#'
#' @noRd
wraplabel <- function(tg) UseMethod("wraplabel")

#' @export
wraplabel.text <- function(tg) {

  label <- getlabel(tg)

  # Split the string into lines
  lines <- unlist(stringi::stri_split(label, regex = "\n"))

  # Identify all the line-wise locations of whitespace in the text (i.e.
  # possible widths to wrap to). Include the full lengths of each line, to
  # allow for non-wrapped wrap
  breakpoints <- lapply(
    lines,
    function(line) stringi::stri_locate_all(line, regex = "\\s")[[1]][,1]
  )
  breakpoints <- c(breakpoints, stringi::stri_length(lines))
  breakpoints <- sort(unique(unlist(breakpoints)))

  # Generate wraps for those lengths
  wraps <- data.frame(wrapwidth = breakpoints)
  wraps$wrap <- vapply(wraps$wrapwidth, function(w) {
    paste0(lapply(lines, function(line) {
      paste0(
        stringi::stri_wrap(str = line, width = w, normalise = FALSE),
        collapse = "\n"
      )
    } ), collapse = "\n")
  }, character(1))

  return(wraps)
}

#' @export
wraplabel.richtext_grob <- function(tg) {

  label <- getlabel(tg)

  # Split the string into lines
  lines <- unlist(strsplit(label, c("<br>", "</br>", "\n")))

  # Identify all the line-wise locations of whitespace in the text (i.e.
  # possible widths to wrap to). Include the full lengths of each line, to
  # allow for non-wrapped wrap
  breakpoints <- lapply(
    lines,
    function(line) stringi::stri_locate_all(line, regex = "\\s")[[1]][,1]
  )
  breakpoints <- c(breakpoints, stringi::stri_length(lines))
  breakpoints <- sort(unique(unlist(breakpoints)))

  # Generate wraps for those lengths
  wraps <- data.frame(wrapwidth = breakpoints)
  wraps$wrap <- vapply(wraps$wrapwidth, function(w) {
    paste0(lapply(lines, function(line) {
      paste0(
        wrap_rich(line, w),
        collapse = "<br>"
      )
    } ), collapse = "<br>")
  }, character(1))

  return(wraps)
}

#' An (approximate) wrapper for strings containing markdown and HTML. 
#'
#' It aims to handle (approximately) the same set of markup as gridtext.
#' Markdown parsing is implemented as a something resembling a finite state
#' machine with memoization (i.e. a packrat parser). The states are as follows:
#'
#' plain: un-marked up text
#' html_start: possibly the start of an HTML tag (i.e. a less-than sign)
#' html: an HTML tag
#' startasterisk1,startasterisk2,startasterisk3: a asterisk belonging to a
#' series of one, two or three that closes a matching pair
#' endasterisk1,endasterisk2,endasterisk3: an asterisk belonging to a series of
#' one, two or three that closes a matching pair
#' markup: other markup characters (that gridtext recognises)
#'
#' @noRd
wrap_rich <- function(string, w) {

  chars <- unlist(strsplit(string, split = NULL))

  # If the string is less than the wrap width, return the string
  if (length(chars) <= w) return(string)

  states <- character(length(chars))
  pos <- 1
  while (pos <= length(chars)) {

    # Skip if this position has already been parsed
    if (! states[pos] == "") {
      pos <- pos + 1; next
    }

    # Make sure the machine hasn't jumped ahead for some reason
    if (pos > 1) {
      if (states[pos - 1] == "") {
        cli::cli_abort("Markdown parser stuck in a bad state at position {pos} - unable to parse")
      }
    }

    # Get character
    char <- chars[pos]

    # Possibly the start of an HTML tag
    if (char == "<") {
      states[pos] <- "html_start"
      pos <- pos + 1; next
    }

    # Possibly the end of an HTML tag. Look back, and if there is an HTML start
    # tag with no intervening HTML, mark the entire tag as HTML. The states of
    # the intervening characters don't matter, a HTML tag steamrolls them all.
    if (char == ">") {
      backpos <- pos - 1
      while (backpos > 0) {
        if (states[backpos] == "html") break
        if (states[backpos] == "html_start") {
          states[seq(backpos, pos)] <- "html"
          break
        }
        backpos <- backpos - 1
      }
      if (states[pos] == "html") {
        pos <- pos + 1; next
      }
    }

    # An asterisk. Look ahead to determine if this is a one, two or
    # three-asterisk series (more than three asterisks don't count, so will begin
    # a new series). Then, look back to see if this is the end of a matched pair.
    if (char == "*") {

      if (chars[pos + 1] == "*" & ! is.na(chars[pos + 1])) {
        if (chars[pos + 2] == "*" & ! is.na(chars[pos + 2])) {
          states[seq(pos, (pos + 2))] <- "asterisk3"
        } else {
          states[seq(pos, (pos + 1))] <- "asterisk2"
        }
      } else {
        states[pos] <- "asterisk1"
      }

      backpos <- pos - 1
      while (backpos > 0) {
        if (states[pos] == "asterisk1") {
          if (states[backpos] == "endasterisk1") break
          if (states[backpos] == "asterisk1") {
            states[backpos] <- "startasterisk1"
            states[pos] <- "endasterisk1"
            break
          }
        }
        if (states[pos] == "asterisk2") {
          if (states[backpos] == "endasterisk2") break
          if (states[backpos] == "asterisk2") {
            states[seq(backpos - 1, backpos)] <- "startasterisk2"
            states[seq(pos, pos + 1)] <- "endasterisk2"
            break
          }
        }
        if (states[pos] == "asterisk3") {
          if (states[backpos] == "endasterisk3") break
          if (states[backpos] == "asterisk3") {
            states[seq(backpos - 2, backpos)] <- "startasterisk3"
            states[seq(pos, pos + 2)] <- "endasterisk3"
            break
          }
        }
        backpos <- backpos - 1
      }

      if (states[pos] == "asterisk1") pos <- pos + 1
      if (states[pos] == "asterisk2") pos <- pos + 2
      if (states[pos] == "asterisk3") pos <- pos + 3
      next
    }

    # Any other markup that gridtext knows about that does not come in a matched
    # pair
    if (char %in% c("^")) {
      states[pos] <- "markup"
      pos <- pos + 1; next
    }

    states[pos] <- "plain"
  }

  # Identify states that count as characters for the purposes of wrapping
  states <- states %in% c("plain", "asterisk1", "asterisk2", "asterisk3")

  # If the number of characters that count is less than the wrap width, return
  # the string
  if (sum(states) <= w) return(string)

  # Wrap to the specified character width
  #
  # States:
  #
  # expand: filling a line out to the desired width
  # contract: having filled to the desired width, backtracking to find somewhere
  # to insert a line break
  # hyperexpand: if there was nowhere within the width window to put a line
  # break, start looking forward to the next possible place (resulting in a
  # too-long line)
  pos <- 1
  line_l <- 0
  state <- "expand"
  while (pos <= length(chars)) {

    if (state == "expand") {
      if (line_l == w) { state <- "contract"; next }
      if (pos == length(chars)) { break }
      if (states[pos]) { line_l <- line_l + 1 }
      pos <- pos + 1
      next
    }

    if (state == "contract") {
      if (chars[pos] == " " & states[pos]) {
        chars[pos] <- "<br>"
        pos <- pos + 1
        line_l <- 0
        state <- "expand"
        next
      }
      if (chars[pos] == "<br>" | pos == 1) {
        state <- "hyperexpand"
        pos <- pos + 1
        next
      }
      pos <- pos - 1
    }

    if (state == "hyperexpand") {
      if (chars[pos] == " " & states[pos]) {
        chars[pos] <- "<br>"
        pos <- pos + 1
        line_l <- 0
        state <- "expand"
        next
      }
      if (pos == length(chars)) break
      pos <- pos + 1
      next
    }
  }
  wrapped_string <- paste(chars, collapse = "")
  return(wrapped_string)
}

theta_rescale <- function(coord, x, panel_params) {
  range <- panel_params$theta.range
  x <- pmin(pmax(x, range[1]), range[2])
  rotate <- function(x) (x + coord$start) %% (2 * pi) * coord$direction
  rotate(
    (x - range[1]) / diff(range) * (2 * pi)
  )
}

r_rescale <- function(coord, x, range) {
  x <- pmin(pmax(x, range[1]), range[2])
  (x - range[1]) / diff(range) * 0.4
}

#' @rdname geom_fit_text
#'
#' @export
geom_bar_text <- function(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  padding.x = grid::unit(1, "mm"),
  padding.y = grid::unit(1, "mm"),
  min.size = 8,
  place = NULL,
  outside = NULL,
  grow = FALSE,
  reflow = FALSE,
  hjust = NULL,
  vjust = NULL,
  fullheight = NULL,
  width = NULL,
  height = NULL,
  formatter = NULL,
  contrast = NULL,
  rich = FALSE,
  ...
) {
  # If position is "identity", outside should be set to TRUE by default;
  # otherwise, set to FALSE
  if (is.null(outside)) {
    outside <- (is.character(position) && position == "identity") ||
      inherits(position, "PositionIdentity")
  }

  ggplot2::layer(
    geom = GeomBarText,
    mapping = mapping,
    data = data,
    stat = stat,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      na.rm = na.rm,
      padding.x = padding.x,
      padding.y = padding.y,
      place = place,
      outside = outside,
      min.size = min.size,
      grow = grow,
      reflow = reflow,
      hjust = hjust,
      vjust = vjust,
      fullheight = fullheight,
      width = width,
      height = height,
      formatter = formatter,
      contrast = contrast,
      rich = rich,
      ...
    )
  )
}

GeomBarText <- ggplot2::ggproto(
  "GeomBarText",
  ggplot2::Geom,
  required_aes = c("x", "y"),
  default_aes = ggplot2::aes(
    label = NULL,
    alpha = 1,
    angle = 0,
    colour = "black",
    family = "",
    fontface = 1,
    lineheight = 0.9,
    size = 12,
    x = NULL,
    y = NULL,
    xmin = NULL,
    xmax = NULL,
    ymin = NULL,
    ymax = NULL
  ),

  setup_data = function(
    data,
    params
  ) {
    # Detect if the bar is to be drawn with 'implied' flipped orientation as
    # permitted in ggplot2 v3.3.0
    implied_flip <- "mapped_discrete" %in% class(data$y)
    data$implied_flip <- implied_flip

    # If the label is missing, assume the value axis is the label (as with
    # stat_count). For flipped bars the value axis is x, otherwise y.
    if (!"label" %in% names(data)) {
      data$label <- if (implied_flip) data$x else data$y
    }

    # Set bar width using the method of geom_boxplot, unless width/height
    # parameters are provided
    if (implied_flip) {
      # If 'height' is provided but not a unit, interpret it as numeric on y scale
      if ((!is.null(params$height)) & (!inherits(params$height, "unit"))) {
        data$ymin <- data$y - params$height / 2
        data$ymax <- data$y + params$height / 2
      } else if (is.null(params$height)) {
        height <- ggplot2::resolution(data$y, FALSE) * 0.9
        data$ymin <- data$y - height / 2
        data$ymax <- data$y + height / 2
      }
    } else {
      # If 'width' is provided but not a unit, interpret it as numeric on x scale
      if ((!is.null(params$width)) & (!inherits(params$width, "unit"))) {
        data$xmin <- data$x - params$width / 2
        data$xmax <- data$x + params$width / 2
      } else if (is.null(params$width)) {
        width <- ggplot2::resolution(data$x, FALSE) * 0.9
        data$xmin <- data$x - width / 2
        data$xmax <- data$x + width / 2
      }
    }

    # Set starting bar height/length
    if (implied_flip) {
      data$xmin <- 0
      data$xmax <- data$x
    } else {
      data$ymin <- 0
      data$ymax <- data$y
    }

    # Apply a formatter function, if one was given
    if (!is.null(params$formatter)) {
      # Check that 'formatter' is a function
      if (!is.function(params$formatter)) {
        cli::cli_abort("{.arg formatter} must be a function")
      }

      # Apply formatter to the labels, checking that the output is a character
      # vector of the correct length
      formatted_labels <- vapply(
        data$label,
        params$formatter,
        character(1),
        USE.NAMES = FALSE
      )
      if (
        (!length(formatted_labels) == length(data$label)) |
          (!is.character(formatted_labels))
      ) {
        cli::cli_abort(
          "{.arg formatter} must produce a character vector of same length as input"
        )
      }
      data$label <- formatted_labels
    }

    data
  },

  draw_key = ggplot2::draw_key_label,

  draw_panel = function(
    data,
    panel_scales,
    coord,
    padding.x = grid::unit(1, "mm"),
    padding.y = grid::unit(1, "mm"),
    min.size = 8,
    place = NULL,
    outside = NULL,
    grow = FALSE,
    reflow = FALSE,
    hjust = NULL,
    vjust = NULL,
    fullheight = NULL,
    width = NULL,
    height = NULL,
    formatter = NULL,
    contrast = TRUE,
    rich = FALSE
  ) {
    # Standardise the place argument
    if (!is.null(place)) {
      if (place %in% c("middle", "center")) {
        place <- "centre"
      }
    }

    # Set contrast if if wasn't set manually. If the text colour is all black,
    # it's probably been left as the default so contrast should be on.
    # Otherwise, it should be turned off
    if (is.null(contrast)) {
      contrast <- all(data$colour == "black", na.rm = TRUE)
    }

    # Detect flipped coordinate system, including the 'implied flip' from
    # ggplot2 v3.3.0
    implied_flip <- data$implied_flip[1]
    is_flipped <- "CoordFlip" %in% class(coord) | implied_flip

    # Set place if it wasn't set manually: top for normal coordinates, right for
    # flipped coordinates
    if (is.null(place)) {
      place <- ifelse(is_flipped, "right", "top")
    }

    # Split data into negative and positive values
    if (implied_flip) {
      positives <- subset(data, data$x >= 0)
      negatives <- subset(data, data$x < 0)
    } else {
      positives <- subset(data, data$y >= 0)
      negatives <- subset(data, data$y < 0)
    }

    # Draw positives with "place" as given
    gtrees <- list()
    if (nrow(positives) > 0) {
      positives <- coord$transform(positives, panel_scales)
      positives_gt <- grid::gTree(
        data = positives,
        padding.x = padding.x,
        padding.y = padding.y,
        place = place,
        outside = outside,
        min.size = min.size,
        grow = grow,
        reflow = reflow,
        width = width,
        height = height,
        contrast = contrast,
        rich = rich,
        cl = "fittexttree"
      )
      gtrees$positives <- positives_gt
    }

    # Draw negatives with reversed place
    if (nrow(negatives) > 0) {
      negatives <- coord$transform(negatives, panel_scales)
      if (place == "top") {
        place <- "bottom"
        negatives$ymax <- negatives$ymin
        negatives$ymin <- negatives$y
      } else if (place == "right") {
        place <- "left"
        negatives$xmax <- negatives$xmin
        negatives$xmin <- negatives$x
      } else if (place == "bottom") {
        place <- "top"
        negatives$ymax <- negatives$ymin
        negatives$ymin <- negatives$y
      } else if (place == "left") {
        place <- "right"
        negatives$xmax <- negatives$xmin
        negatives$xmin <- negatives$x
      }
      negatives_gt <- grid::gTree(
        data = negatives,
        padding.x = padding.x,
        padding.y = padding.y,
        place = place,
        outside = outside,
        min.size = min.size,
        grow = grow,
        reflow = reflow,
        width = width,
        height = height,
        contrast = contrast,
        rich = rich,
        cl = "fittexttree"
      )
      gtrees$negatives <- negatives_gt
    }

    gt <- grid::gTree(children = do.call(grid::gList, gtrees))
    gt$name <- grid::grobName(gt, "geom_bar_text")
    gt
  }
)

#' Some animals and their attributes
#'
#' @format A data frame with 6 rows and 4 variables:
#' \describe{
#'   \item{animal}{the name of the animal}
#'   \item{type}{the type of animal}
#'   \item{flies}{whether the animal can fly}
#'   \item{mass}{the average mass of the animal in grams}
#' }
"animals"

#' Some animals and their attributes, in rich text
#'
#' @format A data frame with 6 rows and 4 variables:
#' \describe{
#'   \item{animal}{the name of the animal, in rich text}
#'   \item{type}{the type of animal}
#'   \item{flies}{whether the animal can fly}
#'   \item{mass}{the average mass of the animal in grams}
#' }
"animals_rich"

#' The compositions of some espresso-based beverages
#'
#' @format A data frame with 6 rows and 3 variables:
#' \describe{
#'   \item{beverage}{the name of the beverage}
#'   \item{ingredient}{the ingredient}
#'   \item{proportion}{the proportion of a cup to fill with the ingredient}
#' }
"beverages"

#' The compositions of some espresso-based beverages, in rich text
#'
#' @format A data frame with 6 rows and 3 variables:
#' \describe{
#'   \item{beverage}{the name of the beverage}
#'   \item{ingredient}{the ingredient, in rich text}
#'   \item{proportion}{the proportion of a cup to fill with the ingredient}
#' }
"beverages_rich"

#' Some craft and their altitudes
#'
#' @format A data frame with 4 rows and 2 variables:
#' \describe{
#'   \item{craft}{the name of the craft}
#'   \item{altitude}{the craft's normal cruising altitude in metres}
#' }
"altitudes"

#' Robert Frost's poem *Nothing Gold Can Stay* (1923)
#'
#' @format A data frame with 8 rows and 6 variables:
#' \describe{
#'   \item{line}{a line from the poem}
#'   \item{xmin}{the xmin coordinate with which to draw the line}
#'   \item{xmax}{the xmax coordinate with which to draw the line}
#'   \item{ymin}{the ymin coordinate with which to draw the line}
#'   \item{ymax}{the ymax coordinate with which to draw the line}
#'   \item{linenumber}{the line number}
#' }
"gold"

## ----fig.width = 4, fig.height = 4--------------------------------------------
library(ggplot2)
library(ggfittext)
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text()

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE, grow = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(place = "topleft", reflow = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(altitudes, aes(x = craft, y = altitude, label = altitude)) +
  geom_col() +
  geom_bar_text()

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(altitudes, aes(x = craft, y = altitude, label = altitude)) +
  geom_col() +
  geom_bar_text(outside = FALSE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(beverages, aes(x = beverage, y = proportion, label = ingredient,
                    fill = ingredient)) +
  geom_col(position = "stack") +
  geom_bar_text(position = "stack", reflow = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(beverages, aes(x = beverage, y = proportion, label = ingredient,
                    fill = ingredient)) +
  geom_col(position = "dodge") +
  geom_bar_text(position = "dodge", grow = TRUE, reflow = TRUE, 
                place = "left") +
  coord_flip()

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(presidential, aes(ymin = start, ymax = end, x = party, label = name)) +
  geom_fit_text(grow = TRUE) +
  geom_errorbar(alpha = 0.5)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals_rich, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE, grow = TRUE, rich = TRUE)

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(gold, aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, 
                 fill = linenumber, label = line)) +
  coord_polar() +
  geom_rect() +
  geom_fit_text(min.size = 0, grow = TRUE) +
  scale_fill_gradient(low = "#fee391", high = "#238443")

## ----fig.width = 4, fig.height = 4--------------------------------------------
ggplot(animals, aes(x = type, y = flies, fill = mass, label = animal)) +
  geom_tile() +
  geom_fit_text(reflow = TRUE, grow = TRUE, contrast = TRUE)

## ----echo = FALSE, fig.width = 4, fig.height = 0.5----------------------------
ggplot(data.frame(label = "fullheight = TRUE", xmin = 0.25, xmax = 0.75, ymin = 0.25, ymax = 0.75), 
       aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = label)) +
  geom_rect() +
  geom_fit_text(grow = TRUE, fullheight = TRUE, family = "mono") +
  theme(axis.text = element_blank(), axis.ticks = element_blank(), panel.background = element_blank(), plot.title = element_text(hjust = 0.5))

## ----echo = FALSE, fig.width = 4, fig.height = 0.5----------------------------
ggplot(data.frame(label = "fullheight = FALSE", xmin = 0.25, xmax = 0.75, ymin = 0.25, ymax = 0.75), 
       aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = label)) +
  geom_rect() +
  geom_fit_text(grow = TRUE, fullheight = FALSE, family = "mono") +
  theme(axis.text = element_blank(), axis.ticks = element_blank(), panel.background = element_blank(), plot.title = element_text(hjust = 0.5))

#' @importFrom grid makeContent
#' @export
makeContent.fittexttreepolar <- function(x) {

  data <- x$data

  # If ymin/ymax are not provided, generate boundary box from height. A similar
  # transformation will be performed for xmin/xmax for each grob individually
  # later, as it needs to be done in the context of the y position of the grob.
  if (!("ymin" %in% names(data))) {
    data$ymin <- data$r - 
      (grid::convertHeight(x$height, "npc", valueOnly = TRUE) / 2)
    data$ymax <- data$r + 
      (grid::convertHeight(x$height, "npc", valueOnly = TRUE) / 2)
  }

  # Handle parameters
  if (is.null(x$contrast)) x$contrast <- FALSE
  if (is.null(x$outside)) x$outside <- FALSE
  if (is.null(x$fullheight)) x$fullheight <- x$grow
  if (x$outside) cli::cli_warn("Outside is not supported in polar coordinates")
  if (x$reflow) cli::cli_warn("Reflowing is not supported in polar coordinates")
  if (! is.null(x$hjust)) cli::cli_warn("hjust is not supported in polar coordinates")
  if (! is.null(x$vjust)) cli::cli_warn("vjust is not supported in polar coordinates")

  # Convert padding.x and padding.y to mm
  padding.x <- grid::convertWidth(x$padding.x, "mm", valueOnly = TRUE)
  padding.y <- grid::convertHeight(x$padding.y, "mm", valueOnly = TRUE)

  # Prepare grob for each text label
  grobs <- lapply(seq_len(nrow(data)), function(i) {

    # Convenience
    text <- data[i, ]

    # Handle angled text
    if (! text$angle == 0) cli::cli_warn("Angled text is not supported in polar coordinates")

    # Set hjust and vjust 
    # A vjust of 0.2 strikes a good visual balance in the kerning of characters
    # in polar coordinates
    x$hjust <- 0.5
    x$vjust <- 0.2

    # Create starting textGrob
    tg <- grid::textGrob(label = text$label, x = 0.5, y = 0.5, default.units = "mm", 
                         hjust = x$hjust, vjust = x$vjust, rot = text$angle, 
                         gp = grid::gpar(col = ggplot2::alpha(text$colour, text$alpha), 
                                         fontsize = text$size, fontfamily = text$family, 
                                         fontface = text$fontface, 
                                         lineheight = text$lineheight))

    # Get starting textGrob dimensions
    tgdim <- tgDimensions(tg, x$fullheight, text$angle)

    # Convert box y coordinates to mm
    ymin <- grid::convertHeight(grid::unit(text$ymin, "npc"), "mm", TRUE)
    ymax <- grid::convertHeight(grid::unit(text$ymax, "npc"), "mm", TRUE)

    # Get dimensions of bounding box. The y dimension will be given in mm, while
    # the x dimension is given as arc length (radians). For convenience of
    # comparing the textGrob to the bounding box on the x dimension, we will
    # also calculate it in mm based on the text placement. If xmin/xmax are
    # not provided, the boundary box will be generated from width.
    ydim <- abs(ymin - ymax) - (2 * padding.y)

    if (!("xmin" %in% names(data))) {
      if (x$place %in% c("bottomleft", "bottom", "bottomright")) {
        r <- ymin + (x$vjust * tgdim$height) + padding.y
      } else if (x$place %in% c("left", "centre", "right")) {
        r <- ((ymin + ymax) / 2) - ((0.5 - x$vjust) * tgdim$height)
      } else if (x$place %in% c("topleft", "top", "topright")) {
        r <- ymax - padding.y - ((1 - x$vjust) * tgdim$height)
      }
      c <- 2 * pi * r
      text$xmin <- text$theta - 
        (((grid::convertWidth(x$width, "mm", valueOnly = TRUE) / 2) / c) * 2 * pi)
      text$xmax <- text$theta + 
        (((grid::convertWidth(x$width, "mm", valueOnly = TRUE) / 2) / c) * 2 * pi)
    }

    xdim <- ifelse(
      text$xmax > text$xmin,
      text$xmax - text$xmin,
      (text$xmax + pi + pi - text$xmin) %% (2 * pi)
    )

    if (x$place %in% c("bottomleft", "bottom", "bottomright")) {
      r <- ymin + (x$vjust * tgdim$height) + padding.y
      xdim_mm <- r * xdim

    } else if (x$place %in% c("left", "centre", "right")) {
      r <- ((ymin + ymax) / 2) - ((0.5 - x$vjust) * tgdim$height)
      xdim_mm <- r * xdim

    } else if (x$place %in% c("topleft", "top", "topright")) {
      r <- ymax - padding.y - ((1 - x$vjust) * tgdim$height)
      xdim_mm <- r * xdim
    }

    # Resize text to fit bounding box if it doesn't fit
    if (
        # Standard condition - is text too big for box?
        (tgdim$width > xdim_mm | tgdim$height > ydim) |
          # grow = TRUE condition - is text too small for box?
          (x$grow & tgdim$width < xdim_mm & tgdim$height < ydim)
        ) {

      # Get the relationships between font size and label dimensions
      slopew <- tg$gp$fontsize / tgdim$width
      slopeh <- tg$gp$fontsize / tgdim$height

      # Calculate the target font size required to make the text fit
      # height-wise
      targetfsh <- ydim * slopeh

      # Calculate the target font size required to make the text fit width-wise
      # See https://imgur.com/a/z5TvFST for explanation of geometry
      if (x$place %in% c("bottomleft", "bottom", "bottomright")) {
        w <- xdim * (ymin + padding.y)
        targetfsw <- w * slopew

      } else if (x$place %in% c("left", "centre", "right")) {
        k <- (tgdim$height * x$vjust) / tgdim$width
        R <- (ymin + ymax) / 2
        w <- ((xdim * R) / ((xdim * k) + 1)) - (2 * padding.x)
        targetfsw <- w * slopew

      } else if (x$place %in% c("topleft", "top", "topright")) {
        k <- tgdim$height / tgdim$width
        R <- ymax - padding.y
        w <- ((xdim * R) / ((xdim * k) + 1)) - (2 * padding.x)
        targetfsw <- w * slopew
      }

      # Set to smaller of target font sizes
      tg$gp$fontsize <- ifelse(targetfsw < targetfsh, targetfsw, targetfsh)
    }

    # Hide if below minimum font size
    if (tg$gp$fontsize < x$min.size) return()

    # Update the textGrob dimensions
    tgdim <- tgDimensions(tg, x$fullheight, text$angle)

    # r = the radius from the centre to the text anchor (which is not the
    # typographic baseline but is defined by vjust). Note that the position of
    # the text anchor does not take descenders into account, so these must be
    # adjusted for if fullheight is true
    if (x$place %in% c("bottomleft", "bottom", "bottomright")) {
      r <- ymin + padding.y + (x$vjust * tgdim$height)
    } else if (x$place %in% c("left", "centre", "right")) {
      r <- ((ymin + ymax) / 2) - ((0.5 - x$vjust) * tgdim$height)
    } else if (x$place %in% c("topleft", "top", "topright")) {
      r <- ymax - padding.y - ((1 - x$vjust) * tgdim$height)
    }
    if (x$fullheight) r <- r + (grid::convertHeight(tgdim$descent, "mm", TRUE) * (1 - x$vjust))

    # c = the circumference of the baseline
    c <- 2 * pi * r

    # char_widths = widths of each character in the string
    chars <- strsplit(as.character(text$label), "")[[1]]
    char_widths <- (grid::calcStringMetric(chars)$width / 
                      sum(grid::calcStringMetric(chars)$width)) * tgdim$width

    # char_arcs = arcwidth of each character, in degrees
    char_arcs <- 360 * char_widths / c

    # padding.x.arcrad = the arcwidth of padding.x, expressed in radians, at
    # the anchor radius
    padding.x.arcrad <- (padding.x / c) * 2 * pi

    # theta = the theta of the text anchor for the entire label in the
    # coordinate system, initial calculated in radians
    if (x$place %in% c("bottomleft", "left", "topleft")) {
      theta <- text$xmin + (deg2rad(sum(char_arcs)) / 2) + padding.x.arcrad
    } else if (x$place %in% c("bottom", "centre", "top")) {
      theta <- ifelse(
        text$xmax > text$xmin,
        (text$xmin + text$xmax) / 2,
        (text$xmin + text$xmax + pi + pi) / 2
      )
    } else if (x$place %in% c("bottomright", "right", "topright")) {
      theta <- text$xmax - (deg2rad(sum(char_arcs)) / 2) - padding.x.arcrad
    }

    # check if need to flip (angle == 180)
    flip <- x$flip

    if (flip) {
      flip <- rad2deg(theta) > 90 && rad2deg(theta) < 270
    }

    # If flipping, over-write necessary strings
    if (flip) {
      # Reverse the string
      text$label <- strrev(as.character(text$label))

      # re-calc string positions
      chars <- strsplit(as.character(text$label), "")[[1]]
      char_widths <- (grid::calcStringMetric(chars)$width / 
                        sum(grid::calcStringMetric(chars)$width)) * tgdim$width

      # char_arcs = arcwidth of each character, in degrees
      char_arcs <- 360 * char_widths / c

      # padding.x.arcrad = the arcwidth of padding.x, expressed in radians, at
      # the anchor radius
      padding.x.arcrad <- (padding.x / c) * 2 * pi

      # theta = the theta of the text anchor for the entire label in the
      # coordinate system, initial calculated in radians
      if (x$place %in% c("bottomleft", "left", "topleft")) {
        theta <- text$xmin + (deg2rad(sum(char_arcs)) / 2) + padding.x.arcrad
      } else if (x$place %in% c("bottom", "centre", "top")) {
        theta <- ifelse(
          text$xmax > text$xmin,
          (text$xmin + text$xmax) / 2,
          (text$xmin + text$xmax + pi + pi) / 2
        )
      } else if (x$place %in% c("bottomright", "right", "topright")) {
        theta <- text$xmax - (deg2rad(sum(char_arcs)) / 2) - padding.x.arcrad
      }
    }


    # angle = ?? I can't even remember what this is supposed to do but it
    # works. Converting from radians to degrees with some sort of correction?
    angle <- 450 - rad2deg(theta)

    # char_thetas = theta position of the anchors for each character (assuming
    # hjust = 0.5 for the textGrob representing this character), in degrees
    lag_vector <- function(x) c(0, x[seq_along(x) - 1])
    char_thetas <- angle - lag_vector(cumsum(char_arcs)) - 
                     (char_arcs / 2) + (sum(char_arcs) / 2)

    # Generate a textGrob for each character
    tgs <- lapply(seq_along(char_thetas), function(i) {

      char <- chars[i]
      theta <- char_thetas[i]
      theta_rad <- deg2rad(theta)

      x_pos <- r * cos(theta_rad)
      x_pos <- 0.5 + grid::convertWidth(grid::unit(x_pos, "mm"), "npc", TRUE)
      y_pos <- r * sin(theta_rad)
      y_pos <- 0.5 + grid::convertHeight(grid::unit(y_pos, "mm"), "npc", TRUE)

      if (flip) {
        # adjust for 0.8 to take into account 0.2 vjust for nice kerning
        x$vjust <- 0.8 - x$vjust
      }

      tg <- grid::textGrob(
        label = char,
        x = x_pos,
        y = y_pos,
        hjust = x$hjust,
        vjust = x$vjust,
        rot = theta - 90 + 180 * flip,
        default.units = "npc",
        gp = grid::gpar(
          fontsize = tg$gp$fontsize,
          col = ggplot2::alpha(text$colour, text$alpha),
          fontfamily = text$family,
          fontface = text$fontface,
          lineheight = text$lineheight
        )
      )
      return(tg)
    })

    # Convert to a gTree
    gt <- grid::gTree(children = do.call(grid::gList, tgs))

    # Return the gTree
    gt
  })

  class(grobs) <- "gList"
  grid::setChildren(x, grobs)
}

---
title: "Introduction to 'ggfittext'"
author: "David Wilkins"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    fig_width: 4
    fig_height: 4
vignette: >
  %\VignetteIndexEntry{Introduction to 'ggfittext'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Fitting text inside a box

Sometimes you want to draw some text in a ggplot2 plot so that it fits inside a
defined area. You can do this by manually fiddling with the font size, but this
is tedious and un-reproducible. ggfittext provides a geom called
`geom_fit_text()` that automatically resizes text to fit inside a box.  It
works like this:

```{r, fig.width = 4, fig.height = 4}
library(ggplot2)
library(ggfittext)
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text()
```

As with `geom_text()`, the position of the text is set by the `x` and `y`
aesthetics. `geom_fit_text()` tries to infer the width and height of the box in
which the text is allowed to fit, and shrinks down any text that is too big.

## Reflowing text

Another way to make the text fit in the box is by reflowing it; that is,
wrapping it over multiple lines. With the `reflow = TRUE` argument,
`geom_fit_text()` will reflow the text before (if still necessary) shrinking
it:

```{r, fig.width = 4, fig.height = 4}
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE)
```

## Growing text

If you want the text to be as large as possible, the argument `grow = TRUE`
will increase the text size to the maximum that will fit in the box. This works
well in conjunction with `reflow`:

```{r, fig.width = 4, fig.height = 4}
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE, grow = TRUE)
```

## Placing text

By default, text is placed in the centre of the box. However, you can place it
in a corner or on a side of the box with the `place` argument, which takes
values like "top", "topright", "bottomleft" and so on:

```{r, fig.width = 4, fig.height = 4}
ggplot(animals, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(place = "topleft", reflow = TRUE)
```

## Bar plots

ggfittext also provides a geom `geom_bar_text()` for labelling bars in bar
plots:

```{r, fig.width = 4, fig.height = 4}
ggplot(altitudes, aes(x = craft, y = altitude, label = altitude)) +
  geom_col() +
  geom_bar_text()
```

`geom_bar_text()` works with stacked bar plots:

```{r, fig.width = 4, fig.height = 4}
ggplot(beverages, aes(x = beverage, y = proportion, label = ingredient,
                    fill = ingredient)) +
  geom_col(position = "stack") +
  geom_bar_text(position = "stack", reflow = TRUE)
```

And it works with dodged bar plots, and with flipped bar plots:

```{r, fig.width = 4, fig.height = 4}
ggplot(beverages, aes(x = beverage, y = proportion, label = ingredient,
                    fill = ingredient)) +
  geom_col(position = "dodge") +
  geom_bar_text(position = "dodge", grow = TRUE, reflow = TRUE, 
                place = "left") +
  coord_flip()
```

## Specifying the box limits

If you want to manually set the limits of the box (instead of having them
inferred from `x` and `y`), you can use `xmin` & `xmax` and/or `ymin` & `ymax`:

```{r, fig.width = 4, fig.height = 4}
ggplot(presidential, aes(ymin = start, ymax = end, x = party, label = name)) +
  geom_fit_text(grow = TRUE) +
  geom_errorbar(alpha = 0.5)
```

Alternatively, you can set the width and/or height with the `width` and/or
`height` arguments, which should be `grid::unit()` objects. The horizontal
and/or vertical centre of the box will be defined by `x` and/or `y`.

### Experimental feature: rich text

With the `rich = TRUE` argument, `geom_fit_text()` and `geom_bar_text()` both
support a limited subset of Markdown and HTML markup for text (rendered with
[gridtext](https://wilkelab.org/gridtext/)).

```{r, fig.width = 4, fig.height = 4}
ggplot(animals_rich, aes(x = type, y = flies, label = animal)) +
  geom_tile(fill = "white", colour = "black") +
  geom_fit_text(reflow = TRUE, grow = TRUE, rich = TRUE)
```

Rich text cannot be drawn in polar coordinates. Please note that this feature
is liable to change, and is subject to upstream changes to gridtext.

## Experimental feature: text in polar coordinates

Text can be drawn in polar coordinates with `geom_fit_text()` simply by adding
`coord_polar()` to the plot. This feature is experimental and any bug reports
are very welcome.

```{r, fig.width = 4, fig.height = 4}
ggplot(gold, aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, 
                 fill = linenumber, label = line)) +
  coord_polar() +
  geom_rect() +
  geom_fit_text(min.size = 0, grow = TRUE) +
  scale_fill_gradient(low = "#fee391", high = "#238443")
```

## Other useful arguments

All arguments to `geom_fit_text()` can also be used with `geom_bar_text()`.

- **`contrast`** can be used to automatically invert the colour of the text so
  it contrasts against a background `fill`:

```{r, fig.width = 4, fig.height = 4}
ggplot(animals, aes(x = type, y = flies, fill = mass, label = animal)) +
  geom_tile() +
  geom_fit_text(reflow = TRUE, grow = TRUE, contrast = TRUE)
```

- **`padding.x`** and **`padding.y`** can be used to set the padding between
  the text and the edge of the box. By default this is 1 mm. These values must
  be given as `grid::unit()` objects.
- **`min.size`** sets the minimum font size in points, by default 4 pt. Text
  smaller than this will be hidden (see also `outside`).
- **`outside`** is `FALSE` by default for `geom_fit_text()`. If `TRUE`, text
  that is placed at "top", "bottom", "left" or "right" and must be shrunk
  smaller than `min.size` to fit in the box will be flipped to the outside of
  the box (if it fits there). This is mostly useful for drawing text inside
  bars in a bar plot.
- **`hjust`** and **`vjust`** set the horizontal and vertical justification of
  the text, scaled between 0 (left/bottom) and 1 (right/top). These are both
  0.5 by default.
- **`formatter`** allows you to provide a function that will be applied to the
  text before it is drawn. This is mostly useful in contexts where variables
  may be interpolated, such as when using
  [gganimate](https://gganimate.com/).
- **`fullheight`** is automatically set depending on place, but can be
  overridden with this option. This is used to determine the bounding box
  around the text. If `FALSE`, the bounding box includes the x-height of the
  text and ascenders, but not any descenders. If TRUE, it extends from the top
  of the ascenders to the bottom of the descenders. This is mostly useful in
  situations where you want to ensure the baseline of text is consistent
  between labels (`fullheight = FALSE`), or when you want to avoid descenders
  spilling out of the bounding box (`fullheight = TRUE`).

```{r echo = FALSE, fig.width = 4, fig.height = 0.5}
ggplot(data.frame(label = "fullheight = TRUE", xmin = 0.25, xmax = 0.75, ymin = 0.25, ymax = 0.75), 
       aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = label)) +
  geom_rect() +
  geom_fit_text(grow = TRUE, fullheight = TRUE, family = "mono") +
  theme(axis.text = element_blank(), axis.ticks = element_blank(), panel.background = element_blank(), plot.title = element_text(hjust = 0.5))
```

```{r echo = FALSE, fig.width = 4, fig.height = 0.5}
ggplot(data.frame(label = "fullheight = FALSE", xmin = 0.25, xmax = 0.75, ymin = 0.25, ymax = 0.75), 
       aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, label = label)) +
  geom_rect() +
  geom_fit_text(grow = TRUE, fullheight = FALSE, family = "mono") +
  theme(axis.text = element_blank(), axis.ticks = element_blank(), panel.background = element_blank(), plot.title = element_text(hjust = 0.5))
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{beverages}
\alias{beverages}
\title{The compositions of some espresso-based beverages}
\format{
A data frame with 6 rows and 3 variables:
\describe{
\item{beverage}{the name of the beverage}
\item{ingredient}{the ingredient}
\item{proportion}{the proportion of a cup to fill with the ingredient}
}
}
\usage{
beverages
}
\description{
The compositions of some espresso-based beverages
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{altitudes}
\alias{altitudes}
\title{Some craft and their altitudes}
\format{
A data frame with 4 rows and 2 variables:
\describe{
\item{craft}{the name of the craft}
\item{altitude}{the craft's normal cruising altitude in metres}
}
}
\usage{
altitudes
}
\description{
Some craft and their altitudes
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{gold}
\alias{gold}
\title{Robert Frost's poem \emph{Nothing Gold Can Stay} (1923)}
\format{
A data frame with 8 rows and 6 variables:
\describe{
\item{line}{a line from the poem}
\item{xmin}{the xmin coordinate with which to draw the line}
\item{xmax}{the xmax coordinate with which to draw the line}
\item{ymin}{the ymin coordinate with which to draw the line}
\item{ymax}{the ymax coordinate with which to draw the line}
\item{linenumber}{the line number}
}
}
\usage{
gold
}
\description{
Robert Frost's poem \emph{Nothing Gold Can Stay} (1923)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom_bar_text.R, R/geom_fit_text.R
\name{geom_bar_text}
\alias{geom_bar_text}
\alias{geom_fit_text}
\title{A 'ggplot2' geom to fit text inside a box}
\usage{
geom_bar_text(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  padding.x = grid::unit(1, "mm"),
  padding.y = grid::unit(1, "mm"),
  min.size = 8,
  place = NULL,
  outside = NULL,
  grow = FALSE,
  reflow = FALSE,
  hjust = NULL,
  vjust = NULL,
  fullheight = NULL,
  width = NULL,
  height = NULL,
  formatter = NULL,
  contrast = NULL,
  rich = FALSE,
  ...
)

geom_fit_text(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  padding.x = grid::unit(1, "mm"),
  padding.y = grid::unit(1, "mm"),
  min.size = 4,
  place = "centre",
  outside = FALSE,
  grow = FALSE,
  reflow = FALSE,
  hjust = NULL,
  vjust = NULL,
  fullheight = NULL,
  width = NULL,
  height = NULL,
  formatter = NULL,
  contrast = FALSE,
  flip = FALSE,
  rich = FALSE,
  ...
)
}
\arguments{
\item{mapping}{\code{ggplot2::aes()} object as standard in 'ggplot2'. Note
that aesthetics specifying the box must be provided. See Details.}

\item{data, stat, position, na.rm, show.legend, inherit.aes, ...}{Standard geom
arguments as for \code{ggplot2::geom_text()}.}

\item{padding.x, padding.y}{Horizontal and vertical padding around the text,
expressed in \code{grid::unit()} objects. Both default to 1 mm.}

\item{min.size}{Minimum font size, in points. Text smaller than this, whether
as supplied or after being shrunk to fit the box, will be hidden. Defaults to
4 pt (8 pt for \code{geom_bar_text()}).}

\item{place}{Where inside the box to place the text. Default is 'centre';
other options are 'topleft', 'top', 'topright', 'right', 'bottomright',
'bottom', 'bottomleft', 'left', and 'center'/'middle' which are both
synonyms for 'centre'. For \code{geom_bar_text()}, will be set heuristically if
not specified.}

\item{outside}{If \code{TRUE}, text placed in one of 'top', 'right', 'bottom' or
'left' that would need to be shrunk smaller than \code{min.size} to fit the box
will be drawn outside the box if possible. This is mostly useful for drawing
text inside bar/column geoms. Defaults to \code{TRUE} for \code{position = "identity"}
when using \code{geom_bar_text()}, otherwise \code{FALSE}.}

\item{grow}{If \code{TRUE}, text will be grown as well as shrunk to fill the box.
Defaults to \code{FALSE}.}

\item{reflow}{If \code{TRUE}, text will be reflowed (wrapped) to better fit the
box. Defaults to \code{FALSE}.}

\item{hjust, vjust}{Horizontal and vertical justification of the text.
\code{vjust} defaults to 0.5. \code{hjust} defaults to 0.5, except for left placements
('left', 'topleft' and 'bottomleft'), where it defaults to 0, and right
placements, where it defaults to 1.}

\item{fullheight}{If \code{TRUE}, descenders will be counted when resizing and
placing text; if \code{FALSE}, only the x-height and ascenders will be counted.
The main use for this option is for aligning text at the baseline (\code{FALSE})
or preventing descenders from spilling outside the box (\code{TRUE}). By default
this is set automatically depending on \code{grow}.}

\item{width, height}{When using \code{x} and/or \code{y} aesthetics, these set the
width and/or height of the box. These should be either \code{grid::unit()}
objects or numeric values on the \code{x} and \code{y} scales.}

\item{formatter}{A function that will be applied to the text before it is
drawn. This is useful when using \code{geom_fit_text()} in contexts involving
interpolated variables, such as with the 'gganimate' package. \code{formatter}
will be applied serially to each element in the \code{label} column, so it does
not need to be a vectorised function.}

\item{contrast}{If \code{TRUE} and in combination with a \code{fill} aesthetic, the
colour of the text will be inverted for better contrast against dark
background fills. \code{FALSE} by default for \code{geom_fit_text()}, set
heuristically for \code{geom_bar_text()}.}

\item{rich}{If \code{TRUE}, text will be formatted with markdown and HTML markup
as implemented by \code{gridtext::richtext_grob()}. \code{FALSE} by default. Rich text
cannot be drawn in polar coordinates. Please note that rich text support is
\strong{experimental} and breaking changes are likely.}

\item{flip}{If \code{TRUE}, when in polar coordinates 'upside-down' text will be
flipped the 'right way up', to enhance readability.}
}
\description{
\code{geom_fit_text()} shrinks, grows and wraps text to fit inside a defined
box.
\code{geom_bar_text()} is a convenience wrapper around \code{geom_fit_text()} for
labelling bar plots generated with \code{geom_col()} and \code{geom_bar()}.
}
\details{
Except where noted, \code{geom_fit_text()} behaves more or less like
\code{ggplot2::geom_text()}.

There are three ways to define the box in which you want the text to be
drawn. The extents of the box on the x and y axes are independent, so any
combination of these methods can be used:
\enumerate{
\item If the \code{x} and/or \code{y} aesthetics are used to set the location of the box,
the width and/or height will be set automatically based on the spacing
between values on the axis, following the method used by
\code{ggplot2::geom_boxplot()}.
\item Alternatively, if \code{x} and/or \code{y} aesthetics are used, the width and/or
height of the box can be overridden with a 'width' and/or 'height' argument.
These should be \code{grid::unit()} objects; if not, they will be assumed to use
the native axis scale.
\item The boundaries of the box can be set using the aesthetics 'xmin' and
'xmax', and/or 'ymin' and 'ymax'.
}

If the text is too big for the box, it will be shrunk to fit the box. With
\code{grow = TRUE}, the text will be made to fill the box completely whether that
requires shrinking or growing.

\code{reflow = TRUE} will cause the text to be reflowed (wrapped) to better fit
in the box. If the text cannot be made to fit by reflowing alone, it will be
reflowed then shrunk to fit the box. Existing line breaks in the text will
be respected when reflowing.

\code{geom_fit_text()} includes experimental support for drawing text in polar
coordinates (by adding \code{coord_polar()} to the plot), however not all
features are available when doing so.
}
\section{Aesthetics}{

\itemize{
\item label (required)
\item (xmin AND xmax) OR x (required)
\item (ymin AND ymax) OR y (required)
\item alpha
\item angle
\item colour
\item family
\item fill (with \code{contrast = TRUE}, used to select a contrasting text colour)
\item fontface
\item lineheight
\item size (in points; note that this differs from \code{ggplot2::geom_text()},
which expresses size in mm)
}

\code{geom_bar_text()} requires only the \code{x} and \code{y} aesthetics, which are used
to infer the extent of the bar. If no \code{label} aesthetic is provided, the
bar's value (\code{y}, or \code{x} for flipped bars) will be used as the label.
}

\examples{

ggplot2::ggplot(ggplot2::presidential, ggplot2::aes(ymin = start, ymax = end,
    label = name, x = party)) +
  geom_fit_text(grow = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggfittext-package.R
\docType{package}
\name{ggfittext-package}
\alias{ggfittext}
\alias{ggfittext-package}
\title{ggfittext: Fit Text Inside a Box in 'ggplot2'}
\description{
A 'ggplot2' extension to fit text into a box by growing, shrinking or wrapping the text.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://wilkox.org/ggfittext/}
  \item Report bugs at \url{https://github.com/wilkox/ggfittext/issues/}
}

}
\author{
\strong{Maintainer}: David Wilkins \email{david@wilkox.org}

Authors:
\itemize{
  \item David Wilkins \email{david@wilkox.org}
}

Other contributors:
\itemize{
  \item Brady Johnston \email{brady.johnston@me.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{beverages_rich}
\alias{beverages_rich}
\title{The compositions of some espresso-based beverages, in rich text}
\format{
A data frame with 6 rows and 3 variables:
\describe{
\item{beverage}{the name of the beverage}
\item{ingredient}{the ingredient, in rich text}
\item{proportion}{the proportion of a cup to fill with the ingredient}
}
}
\usage{
beverages_rich
}
\description{
The compositions of some espresso-based beverages, in rich text
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{animals_rich}
\alias{animals_rich}
\title{Some animals and their attributes, in rich text}
\format{
A data frame with 6 rows and 4 variables:
\describe{
\item{animal}{the name of the animal, in rich text}
\item{type}{the type of animal}
\item{flies}{whether the animal can fly}
\item{mass}{the average mass of the animal in grams}
}
}
\usage{
animals_rich
}
\description{
Some animals and their attributes, in rich text
}
\keyword{datasets}
